from __future__ import annotations

import dataclasses
from typing import Callable

from icclim.models.climate_variable import ClimateVariable
from icclim.models.frequency import Frequency
from icclim.models.logical_link import LogicalLink
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation


@dataclasses.dataclass
class IndexConfig:
    """
    DTO class to map icclim.index input the parameters of the different indicator
    compute functions.

    Attributes
    ----------
    frequency: Frequency
        The expected resampling frequency of the output.
    climate_variables: list[ClimateVariable]
        List of CfVariable necessary to compute the index.
    save_percentile: bool = False
        On percentile based indices, if True, this saves the percentile in the output
        netcdf.
    netcdf_version:
        Netcdf version to be used when creating the output
    rolling_window_width:
        On indices relying on a rolling window of days, configure the window width.
    scalar_thresholds:
        On indices relying on a threshold, configure the threshold value. Unit less.
        The unit "degC" is added by icclim.
    transfer_limit_Mbytes:
        The dask maximum chunk size.
    out_unit:
        The output unit overriding Xclim results.
    callback:
        A callable to produce a progress bar
    """

    frequency: Frequency
    climate_variables: list[ClimateVariable]
    min_spell_length: int | None
    rolling_window_width: int | None
    out_unit: str | None
    callback: Callable[[int], None] | None
    netcdf_version: NetcdfVersion
    save_thresholds: bool
    interpolation: QuantileInterpolation
    is_compared_to_reference: bool
    reference_period: tuple[str, str] | None
    indicator_name: str
    logical_link: LogicalLink
    coef: float | None
    date_event: bool
    sampling_method: str



from __future__ import annotations

from typing import TypedDict

# fmt: off
# flake8: noqa

class IndicatorMetadata(TypedDict):
    standard_name: str
    long_name: str
    cell_methods: str

COMBINED_VARS_LONG_NAME = (
    "{% for i, climate_var in enumerate(climate_vars) %}"
        "{{climate_var.long_name}} is"
        " {{climate_var.threshold.long_name}}"
        "{% if i != len(climate_vars) - 1 %}"
            " And "
        "{% endif%}"
    "{% endfor %}"
    " for each {{output_freq.long_name}}."
)
COMBINED_VARS_STANDARD_NAME = (
    "{% for i, climate_var in enumerate(climate_vars) %}"
        "{{climate_var.standard_name}}"
        "_is_{{climate_var.threshold.standard_name}}"
        "{% if i != len(climate_vars) - 1 %}"
            "_and_"
        "{% endif %}"
    "{% endfor %}"
)
SINGLE_VAR_LONG_NAME = (
    "{{source_freq.adjective}}"
    " {{climate_vars[0].long_name}}"
    " related to {{climate_vars[0].threshold.value}}"
    " for each {{output_freq.long_name}}."
)
SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE = (
    "{{source_freq.adjective}}"
    " {{climate_vars[0].long_name}}"
    "{% if climate_vars[0].threshold %}"
        " when {{climate_vars[0].long_name}}"
        " is {{climate_vars[0].threshold.long_name}}"
    "{% endif %}"
    " for each {{output_freq.long_name}}."
)

INDICATORS_TEMPLATES_EN: dict[str, IndicatorMetadata] = {
    "count_occurrences": {
        "long_name":     "Number of {{source_freq.units}}"
                         f" when {COMBINED_VARS_LONG_NAME}",
        "standard_name": "number_of_{{source_freq.units}}_when"
                         f"_{COMBINED_VARS_STANDARD_NAME}",
        "cell_methods":  "time: sum over {{source_freq.units}}",
    },
    "max_consecutive_occurrence": {
        "standard_name": "spell_length_of_{{source_freq.units}}_with"
                         f"_{COMBINED_VARS_STANDARD_NAME}",
        "long_name":     "Maximum number of consecutive {{source_freq.units}} when"
                         f" {COMBINED_VARS_LONG_NAME}",
        "cell_methods":  "time: maximum over {{source_freq.units}}",
    },
    "sum_of_spell_lengths": {
        "standard_name":  # not CF
                        "spell_length_of_{{source_freq.units}}_with"
                        f"_{COMBINED_VARS_STANDARD_NAME}",
        "long_name":    "Sum of spell lengths of at least {{min_spell_length}}"
                        " {{source_freq.units}} when"
                        f" {COMBINED_VARS_LONG_NAME}",
        "cell_methods": "time: sum over {{source_freq.units}}",
    },
    "excess": {
        "standard_name": "integral_of"
                         "_{{climate_vars[0].standard_name}}"
                         "_excess_wrt_time",
        "long_name":     f"Excess of {SINGLE_VAR_LONG_NAME}",
        "cell_methods":  "time: difference over {{source_freq.units}}",
    },
    "deficit": {
        "standard_name": "integral_of_{{climate_vars[0].standard_name}}"
                         "_deficit_wrt_time",
        "long_name":     f"Deficit of {SINGLE_VAR_LONG_NAME}",
        "cell_methods":  "time: difference over {{source_freq.units}}",
    },
    "fraction_of_total": {
        "standard_name":  # not CF
                        "fraction_of_thresholded_{{climate_vars[0].standard_name}}"
                        "_on_total",
        "long_name":    f"Fraction of {SINGLE_VAR_LONG_NAME}",
        # not cf
        "cell_methods": "time: fraction over {{source_freq.units}}",
    },
    "maximum": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     f"Maximum of {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: maximum over {{source_freq.units}}",
    },
    "minimum": {
        "standard_name":"{{climate_vars[0].standard_name}}",
        "long_name":    "Minimum of"
                        f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods": "time: minimum over {{source_freq.units}}",
    },
    "average": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Average of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: mean over {{source_freq.units}}",
    },
    "sum": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Sum of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: sum over {{source_freq.units}}",
    },
    "standard_deviation": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Standard deviation of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: standard_deviation over {{source_freq.units}}",
    },
    "max_of_rolling_sum": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Maximum {{rolling_window_width}}"
                         " {{source_freq.units}} rolling sum of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: sum over {{source_freq.units}}",
    },
    "min_of_rolling_sum": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Minimum {{rolling_window_width}}"
                         " {{source_freq.units}} rolling sum of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: sum over {{source_freq.units}}",
    },
    "min_of_rolling_average": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Minimum {{rolling_window_width}}"
                         " {{source_freq.units}} rolling average of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: mean over {{source_freq.units}}",
    },
    "max_of_rolling_average": {
        "standard_name": "{{climate_vars[0].standard_name}}",
        "long_name":     "Maximum {{rolling_window_width}}"
                         " {{source_freq.units}} rolling average of"
                         f" {SINGLE_VAR_LONG_NAME_WITH_EXCEEDANCE}",
        "cell_methods":  "time: mean over {{source_freq.units}}",
    },
    "mean_of_difference": {
        "standard_name": "range_between_{{climate_vars[0].standard_name}}"
                         "_and_{{climate_vars[1].standard_name}}", # not CF
        "long_name":     "Mean of difference between {{climate_vars[0].long_name}}"
                         " and {{climate_vars[1].long_name}}"
                         " for each {{output_freq.long_name}}.",
        "cell_methods":  "time: range within {{source_freq.units}}"
                         " time: mean over {{source_freq.units}}",
    },
    "difference_of_extremes":  {
        "standard_name": "range_of_extremes_between_{{climate_vars[0].standard_name}}"
                         "_and_{{climate_vars[1].standard_name}}", # not CF
        "long_name":     "Difference between"
                         " maximum of {{source_freq.adjective}}"
                         " {{climate_vars[0].long_name}}"
                         " and minimum of {{source_freq.adjective}}"
                         " {{climate_vars[1].long_name}}"
                         " for each {{output_freq.long_name}}.",
        "cell_methods":  "time: range within {{source_freq.units}}"
                         " time: maximum over {{source_freq.units}}"
                         " time: minimum over {{source_freq.units}}",
    },
    "mean_of_absolute_one_time_step_difference":  {
        "standard_name": "variability_range_between_{{climate_vars[0].standard_name}}"
                         "_and_{{climate_vars[1].standard_name}}", # not CF
        "long_name":     "Average of the absolute {{source_freq.long_name}}"
                         " to {{source_freq.long_name}} difference"
                         " of the {{source_freq.adjective}} variation between"
                         " {{climate_vars[0].long_name}}"
                         " and {{climate_vars[1].long_name}}"
                         " for each {{output_freq.long_name}}.",
        "cell_methods":  "time: range within {{source_freq.units}}"
                         " time: difference over {{source_freq.units}}"
                         " time: mean over {{source_freq.units}}",
    },
    "difference_of_means":  {
        "standard_name": "{{climate_vars[0].standard_name}}"
                         "{% if not is_compared_to_reference %}"
                            "_to_{{climate_vars[1].standard_name}}"
                         "{% endif%}"
                         "_anomaly", # not CF
        "long_name":     "{{output_freq.adjective}} difference between the"
                         " averaged {{source_freq.adjective}}"
                         " {{climate_vars[0].long_name}}"
                         " and"
                         "{% if is_compared_to_reference %}"
                            " its averaged {{source_freq.adjective}} values for the"
                            " {{reference_period}} period."
                         "{% else %}"
                            " the {{output_freq.adjective}}"
                            " averaged {{source_freq.adjective}}"
                            " {{climate_vars[1].long_name}}"
                         "{% endif%}",
        "cell_methods":  "time: mean over {{source_freq.units}}"
                         " time: difference over {{source_freq.units}}",
    },
}



from icclim._generated_api import *  # noqa
from icclim.generic_indices.threshold import build_threshold
from icclim.main import index, indice, indices
from icclim.models.constants import ICCLIM_VERSION
from icclim.pre_processing.rechunk import create_optimized_zarr_store

__all__ = [
    # -- Base functions:
    "index",
    "indice",  # deprecated
    "indices",
    # -- Rechunk function:
    "create_optimized_zarr_store",
    # -- Threshold factory function
    "build_threshold",
]

__version__ = ICCLIM_VERSION



from __future__ import annotations

import cftime
import numpy as np
import pandas as pd
import pytest

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.frequency import FrequencyRegistry, get_seasonal_time_updater
from icclim.tests.testing_utils import stub_tas


class Test_build_frequency_over_frequency:
    def test_simple(self):
        freq = FrequencyRegistry.lookup(FrequencyRegistry.YEAR)
        assert freq == FrequencyRegistry.YEAR


class Test_build_frequency_over_string:
    def test_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            FrequencyRegistry.lookup("yolo")

    def test_simple(self):
        freq = FrequencyRegistry.lookup("year")
        assert freq == FrequencyRegistry.YEAR


class Test_build_frequency_over_list:
    def test_lookup_list__keyword_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            FrequencyRegistry.lookup(["cacahuêtes"])

    def test_lookup_string_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            FrequencyRegistry.lookup("cacahuêtes")

    def test_lookup_month(self):
        freq = FrequencyRegistry.lookup(["month", [1, 4, 3]])
        assert freq.pandas_freq == "MS"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_lookup_season(self):
        freq = FrequencyRegistry.lookup(["season", [1, 2, 3, 4]])
        assert freq.pandas_freq == "AS-JAN"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_lookup_season_tuple(self):
        freq = FrequencyRegistry.lookup(("season", [1, 2, 3, 4]))
        assert freq.pandas_freq == "AS-JAN"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_lookup_pandas_freq(self):
        freq = FrequencyRegistry.lookup("3MS")
        assert freq.pandas_freq == "3MS"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_lookup_winter__deprecated_tuple(self):
        freq = FrequencyRegistry.lookup(["season", ([11, 12], [1, 2, 3, 4])])
        assert freq.pandas_freq == "AS-NOV"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_lookup_error__non_consecutive_season(self):
        with pytest.raises(InvalidIcclimArgumentError):
            FrequencyRegistry.lookup(["season", ([12, 3])])

    def test_lookup_error__weird_months(self):
        with pytest.raises(InvalidIcclimArgumentError):
            FrequencyRegistry.lookup(["season", ([42, 0])])

    def test_lookup__winter(self):
        freq = FrequencyRegistry.lookup(["season", [11, 12, 1, 2]])
        assert freq.pandas_freq == "AS-NOV"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_lookup_season__between_dates(self):
        freq = FrequencyRegistry.lookup(["season", ["07-19", "08-14"]])
        assert freq.pandas_freq == "AS-JUL"
        assert freq.accepted_values == []
        assert freq.post_processing is not None


class Test_seasons_resampler:
    def test_simple(self):
        # WHEN
        test_da = filter_months(stub_tas(), [4, 5, 6]).resample(time="YS").mean()
        da_res, time_bds_res = get_seasonal_time_updater(4, 6)(test_da)
        # THEN
        np.testing.assert_array_equal(1, da_res)
        assert time_bds_res[0].data[0] == pd.to_datetime("2042-04")
        assert (
            time_bds_res[0].data[1]
            == pd.to_datetime("2042-07") - pd.tseries.offsets.Day()
        )

    def test_winter(self):
        # WHEN
        test_da = filter_months(stub_tas(), [11, 12, 1]).resample(time="AS-NOV").mean()
        da_res, time_bds_res = get_seasonal_time_updater(11, 1)(test_da)
        # THEN
        np.testing.assert_array_equal(1, da_res)
        assert time_bds_res[0].data[0] == pd.to_datetime("2041-11")
        assert (
            time_bds_res[0].data[1]
            == pd.to_datetime("2042-02") - pd.tseries.offsets.Day()
        )
        assert time_bds_res[1].data[0] == pd.to_datetime("2042-11")
        assert (
            time_bds_res[1].data[1]
            == pd.to_datetime("2043-02") - pd.tseries.offsets.Day()
        )

    @pytest.mark.parametrize("use_cf", [True, False])
    def test_between_dates(self, use_cf):
        # WHEN
        test_da = (
            filter_months(stub_tas(use_cftime=use_cf), [11, 12, 1])
            .resample(time="AS-NOV")
            .mean()
        )
        da_res, time_bds_res = get_seasonal_time_updater(11, 1, 2, 30)(test_da)
        # THEN
        np.testing.assert_array_equal(1, da_res)  # data must be unchanged
        assert time_bds_res[0].data[0] == pd.to_datetime("2041-11-02")
        assert time_bds_res[0].data[1] == pd.to_datetime("2042-01-30")

    @pytest.mark.parametrize("use_cf", [True, False])
    def test_between_dates__december_ending(self, use_cf):
        # WHEN
        test_da = (
            filter_months(stub_tas(use_cftime=use_cf), [11, 12])
            .resample(time="AS-NOV")
            .mean()
        )
        da_res, time_bds_res = get_seasonal_time_updater(
            start_month=11,
            end_month=12,
            start_day=2,
        )(test_da)
        # THEN
        if use_cf:
            assert da_res.time[0] == cftime.DatetimeGregorian(2042, 12, 1, 12)
        else:
            assert da_res.time[0] == pd.to_datetime("2042-12-01 12:00")
        assert time_bds_res[0].data[0] == pd.to_datetime("2042-11-02")
        assert time_bds_res[0].data[1] == pd.to_datetime("2042-12-31")


def filter_months(da, month_list: list[int]):
    return da.sel(time=da.time.dt.month.isin(month_list))



from __future__ import annotations

import abc
import re
from datetime import datetime
from typing import Any, Callable, Sequence, TypedDict, Union

import jinja2
import numpy as np
import pint
import xarray as xr
from xarray import DataArray, Dataset
from xclim.core.bootstrapping import percentile_bootstrap
from xclim.core.calendar import build_climatology_bounds, percentile_doy, resample_doy
from xclim.core.units import convert_units_to, str2pint
from xclim.core.units import units as xc_units
from xclim.core.utils import PercentileDataArray, calc_perc

from icclim.generic_indices.threshold_templates import (
    EN_THRESHOLD_TEMPLATE,
    PercentileTemplateConfig,
    ThresholdMetadata,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import (
    DEFAULT_DOY_WINDOW,
    DOY_COORDINATE,
    DOY_PERCENTILE_UNIT,
    PERIOD_PERCENTILE_UNIT,
    UNITS_KEY,
)
from icclim.models.frequency import Frequency
from icclim.models.logical_link import LogicalLink, LogicalLinkRegistry
from icclim.models.operator import Operator, OperatorRegistry
from icclim.models.quantile_interpolation import (
    QuantileInterpolation,
    QuantileInterpolationRegistry,
)
from icclim.pre_processing.input_parsing import (
    build_reference_da,
    find_standard_vars,
    get_name_of_first_var,
    is_dataset_path,
    read_clim_bounds,
    read_dataset,
    standardize_percentile_dim_name,
)
from icclim.utils import is_number_sequence

ThresholdValueType = Union[
    str, float, int, Dataset, DataArray, Sequence[Union[float, int, str]], None
]


class ThresholdBuilderInput(TypedDict, total=False):
    """
    Data transfert object mapping all possible configuration to build any threshold.
    """

    operator: Operator
    value: ThresholdValueType
    unit: str | None
    threshold_var_name: str | None
    initial_query: str | None
    threshold_min_value: pint.Quantity | None
    # percentile conf:
    doy_window_width: int | None
    only_leap_years: bool | None
    interpolation: str | QuantileInterpolation | None
    reference_period: Sequence[datetime | str] | None
    # bounded conf:
    thresholds: tuple[
        ThresholdBuilderInput | Threshold, ThresholdBuilderInput | Threshold
    ] | None
    logical_link: LogicalLink


def build_threshold(
    query: str | None = None,
    *,
    operator: Operator | str = None,
    value: ThresholdValueType = None,
    unit: str | None = None,
    threshold_min_value: str | float | None | pint.Quantity = None,
    thresholds: Sequence[Threshold | str, Threshold | str] | None = None,
    logical_link: str | LogicalLink | None = None,
    **kwargs,
) -> BoundedThreshold | PercentileThreshold | BasicThreshold:
    """Factory for thresholds.

    Parameters
    ----------

    query: str | None = None
        string query describing a threshold.
        It must include: an operator, a threshold value and optionally a unit
        such as "> 10 degC".
        It acts as a shorthand for ``operator``, ``value`` and ``unit`` parameters for
        simple threshold.
        Don't use ``query`` when value is a DataArray, a Dataset or a path to a
        netcdf/zarr storage, instead use ``operator``, ``value`` and ``unit``.
        query supersede `operator`, `value` and `unit` parameters.
    operator: Operator | str = None
        keyword argument only.
        The operator either as an instance of Operator or as a compatible string.
        See :py:class:`OperatorRegistry` for the list of all operators.
        When query is None and operator is None, the default ``Operator.REACH`` is used.
    value: str | float | int | Dataset | DataArray | Sequence[float | int | str] | None
        keyword argument only.
        The threshold value(s), default to None.
        It can be:
        * a simple scalar threshold
        * a percentile that will be computed per-grid cell (in combinaison with `unit`)
        * per-grid cell thresholds defined by a DataArray, a Dataset or a string path to
        a netcdf/zarr.
        * a sequence of scalars, the indicator will then be computed for each value and
        a specific dimension will be created (also work with percentiles).
    unit: str | None = None
        Keyword argument only.
        The threshold unit.
        When ``unit`` is None, if ``value`` is a dataset and a "units"
        can be read from its `attrs`, this unit will be used. If value is a scalar or
        a sequence of scalar, the exceedance will be computed by assuming threshold has
        the same unit as the studied value is it compared to.
        When unit is a string it must be a valid unit of our shared pint registry with
        xclim or a special percentile unit:
        * "doy_per" for day of year percentiles (in ECAD, used for temperature indices
        such as TX90p)
        * "period_per" for per period percentiles (in ECAD, used for rain indices such
        as R75p)
    threshold_min_value: str | float | pint.Quantity
        A minimum value used to pre-filter computed threshold values.
        This is particularly useful to compute a percentile threshold for rain where
        dry days are usually ignored. In that case threshold_min_value would be set to
        "1 mm/day".
        If threshold_min_value is a number, ``unit`` is used to quantify
        ``threshold_min_value``.
    kwargs
        Additional arguments to build a PercentileThreshold.
        See :py:class:`PercentileThreshold` constructor for the complete list
        of possible arguments.

    Examples
    --------
    .. code-block:: python

        # -- Scalar threshold
        scalar_t = build_threshold(">= 30 degC")
        assert isinstance(scalar_t, BasicThreshold)

        # -- Daily percentile threshold
        doy_t = build_threshold(">= 30 doy_per")
        assert isinstance(doy_t, PercentileThreshold)

        # -- Per grid-cell threshold
        grided_t = build_threshold(
            operator=">=", value="path/to/tasmax_thresholds.nc", unit="K"
        )
        assert isinstance(grided_t, BasicThreshold)

        # -- Daily percentile threshold, from a file
        tasmax = xarray.open_dataset("path/to/tasmax_thresholds.nc").tasmax
        doys = xclim.core.calendar.percentile_doy(tasmax)
        doy_file_t = build_threshold(operator=">=", value=doys)
        assert isinstance(doy_file_t, PercentileThreshold)

        # -- Bounded threshold
        bounded_t = build_threshold(">= -20 degree AND <= 20 degree ")
        # equivalent to:
        x = build_threshold(">= -20 degree")
        y = build_threshold("<= 20 degree")
        bounded_t2 = x & y
        assert bounded_t == bounded_t2
        # equivalent to:
        bounded_t3 = build_threshold(thresholds=[x, y], logical_link="AND")
        assert bounded_t == bounded_t3
        assert isinstance(bounded_t, BoundedThreshold)


    """
    input_thresh = _read_input(
        query,
        operator,
        value,
        unit,
        threshold_min_value,
        thresholds,
        logical_link,
        **kwargs,
    )
    if _must_build_per_threshold(input_thresh):
        return PercentileThreshold(**input_thresh)
    elif _must_build_basic_threshold(input_thresh):
        return BasicThreshold(**input_thresh)
    elif _must_build_bounded_threshold(input_thresh):
        return BoundedThreshold(**input_thresh)
    else:
        raise NotImplementedError(f"Threshold cannot be built from a {type(value)}")


class Threshold(metaclass=abc.ABCMeta):
    """
    Abstract super class for all thresholds.
    """

    operator: Operator | str
    value: ThresholdValueType
    unit: str | None = None
    initial_query: str | None
    threshold_min_value: pint.Quantity | None = None
    threshold_var_name: str | None

    @abc.abstractmethod
    def format_metadata(
        self, *, jinja_scope: dict[str, Any], jinja_env: jinja2.Environment, **kwargs
    ) -> ThresholdMetadata:
        """Get a dictionary of standardized threshold metadata."""
        ...

    @abc.abstractmethod
    def __eq__(self, other: Threshold) -> bool:
        ...

    def __and__(self, other: Threshold) -> BoundedThreshold:
        """
        Shorthand to build a BoundedThreshold from two existing Threshold with a
        "AND" LogicalLink.
        """
        return BoundedThreshold(
            thresholds=[self, other],
            logical_link=LogicalLinkRegistry.LOGICAL_AND,
            initial_query=None,
        )

    def __or__(self, other: Threshold) -> BoundedThreshold:
        """
        Shorthand to build a BoundedThreshold from two existing Threshold with a
        "OR" LogicalLink.
        """
        return BoundedThreshold(
            thresholds=[self, other],
            logical_link=LogicalLinkRegistry.LOGICAL_OR,
            initial_query=None,
        )

    @abc.abstractmethod
    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,
    ) -> DataArray:
        """Compute the exceedance of comparison_data over/below self threshold
        depending on `self` operator.
        The operator can be overridden by `override_op`. This is needed when
        self.operator is REACH.

        Parameters
        ----------
        comparison_data: xr.DataArray
            Data that must be compared to self threshold
        override_op: Callable[[DataArray, DataArray], DataArray] | None
            Operator to override self.operator compute function.
            Optional.
        kwargs:
            Keyword arguments passed to the concrete compute implementations
            This makes the `compute` interface contract not always reliable.
            So we assume to not really respected LSP here.
        """
        ...


class BoundedThreshold(Threshold):
    """
    Threshold that binds two other thresholds (e.g. "> 95 doy_per AND >= 30 deg_C").

    The logical link must be either "and" or "or".
    """

    left_threshold: Threshold
    right_threshold: Threshold
    logical_link: LogicalLink

    @property
    def unit(self) -> str | None:
        if self.left_threshold.unit == self.right_threshold.unit:
            return self.left_threshold.unit
        else:
            return None

    @unit.setter
    def unit(self, unit):
        self.left_threshold.unit = unit
        self.right_threshold.unit = unit

    def __init__(
        self,
        thresholds: Sequence[Threshold | str | ThresholdBuilderInput],
        logical_link: LogicalLink,
        initial_query: str | None,
        **kwargs,  # noqa
    ):
        if len(thresholds) != 2:
            raise InvalidIcclimArgumentError(
                f"BoundedThreshold can only be built on 2 thresholds, {len(thresholds)}"
                f" were found."
            )
        self.left_threshold = self._build_thresh(thresholds[0])
        self.right_threshold = self._build_thresh(thresholds[1])
        if self.left_threshold == self.right_threshold:
            raise InvalidIcclimArgumentError(
                f"BoundedThreshold must be built on 2 **different** thresholds, here"
                f" both were {self.left_threshold.initial_query}"
            )
        self.logical_link = logical_link
        self.initial_query = initial_query

    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,
    ) -> DataArray:
        left_res = self.left_threshold.compute(
            comparison_data, override_op=override_op, **kwargs
        )
        right_res = self.right_threshold.compute(
            comparison_data, override_op=override_op, **kwargs
        )
        return self.logical_link.compute([left_res, right_res])

    def format_metadata(
        self, *, jinja_scope: dict[str, Any], jinja_env: jinja2.Environment, **kwargs
    ) -> ThresholdMetadata:
        templates = self._get_metadata_templates()
        conf = {
            "left_threshold": self.left_threshold.format_metadata(
                jinja_scope=jinja_scope, jinja_env=jinja_env, **kwargs
            ),
            "logical_link": self.logical_link,
            "right_threshold": self.right_threshold.format_metadata(
                jinja_scope=jinja_scope, jinja_env=jinja_env, **kwargs
            ),
        }
        conf.update(jinja_scope)
        return {
            k: jinja_env.from_string(v, globals=conf).render()
            for k, v in templates.items()
        }

    def __eq__(self, other: BoundedThreshold) -> bool:
        """
        BoundedThreshold equality is based on the reflexive property of logical links (
        and/or). Thus, the left_threshold of `self` can be either equal to the
        left_threshold or the right_threshold of `other`.
        Same goes for right_threshold of self.

        Parameters
        ----------
        other : BoundedThreshold
            The comparison threshold.

        Returns
        -------
            True if the the comparison threshold `other` is equivalent to `self`.
        """
        return (
            isinstance(other, BoundedThreshold)
            and self.initial_query == other.initial_query
            and (
                self.left_threshold == other.left_threshold
                or self.left_threshold == other.right_threshold
            )
            and (
                self.right_threshold == other.left_threshold
                or self.right_threshold == other.right_threshold
            )
            and self.logical_link == other.logical_link
        )

    def _build_thresh(
        self, thresh_input: Threshold | str | ThresholdBuilderInput
    ) -> Threshold:
        if isinstance(thresh_input, Threshold):
            return thresh_input
        elif isinstance(thresh_input, str):
            return build_threshold(thresh_input)
        else:
            return build_threshold(**thresh_input)

    def _get_metadata_templates(self) -> ThresholdMetadata:
        return EN_THRESHOLD_TEMPLATE["bounded_threshold"]


class PercentileThreshold(Threshold):
    """
    Percentile based threshold (e.g. "<= 10 doy_per").

    The percentiles to be computed are expected to be either:

    * "doy percentiles" (unit: "doy_per"). They are usually used for temperatures
      indices such as the ECAD :ref:`tx90p <ecad_functions_api>`.
      These percentiles are computed per day of year (doy) and by aggregating
      values on the time axis ranged by ``reference_period``, using the
      ``doy_window_width`` parameter to control the time axis window of aggregation.
      The resulting `value` is a DataArray with a "dayofyear" dimension ranging from
      0 to 365 with one value per day of the year.
    * "period percentiles" (unit: "period_per"). They are usually used for liquide
      precipitation indices such as the ECAD :ref:`r75p <ecad_functions_api>`
      or even :ref:`r75ptot <ecad_functions_api>`.
      These percentiles are computed per grid cell on the period ranged by
      ``reference_period``.
      The resulting ``value`` is a DataArray with per grid cell values and no time axis.

    ``is_ready`` becomes True when `prepare` method has been called, the actual
    percentiles are then computed and accessible in ``value`` property.
    Once ``is_ready`` is True, ``unit`` property can be set and will attempt a pint unit
    conversion similar to what is done on ``BasicThreshold``.
    Before that, setting unit has no effect.
    """

    reference_period: Sequence[str]
    doy_window_width: int
    only_leap_years: bool
    interpolation: QuantileInterpolation

    is_doy_per_threshold: bool
    _prepared_value: PercentileDataArray
    _initial_unit: str | None
    _initial_value: list[float] | None

    @property
    def unit(self) -> str | None:
        if self.is_ready:
            return self._prepared_value.attrs[UNITS_KEY]
        return self._initial_unit

    @unit.setter
    def unit(self, unit: str | xr.DataArray | pint.Quantity | pint.Unit):
        if self.is_ready:
            if self.value.attrs.get(UNITS_KEY, None) is not None and unit is not None:
                self._prepared_value = convert_units_to(self._prepared_value, unit)
            self.value.attrs[UNITS_KEY] = unit

    @property
    def value(self) -> PercentileDataArray:
        if self.is_ready:
            return self._prepared_value
        else:
            raise RuntimeError(
                "Property `value` is not ready. For PercentileDataArray,"
                " you must call `.prepare` first and fill `studied_data`"
                " parameter in order to prepare `value`."
            )

    def __init__(
        self,
        operator: str | Operator,
        value: DataArray | float | int | Sequence[float],
        unit: str | None = None,
        doy_window_width: int = DEFAULT_DOY_WINDOW,
        only_leap_years: bool = False,
        interpolation: QuantileInterpolation
        | str = QuantileInterpolationRegistry.MEDIAN_UNBIASED,
        reference_period: Sequence[datetime | str] | None = None,
        threshold_min_value: pint.Quantity | None = None,
        initial_query: str | None = None,
        threshold_var_name: str | None = None,
        **kwargs,  # noqa
    ):
        if is_dataset_path(value) or isinstance(value, Dataset):
            value, is_doy_per_threshold = _build_per_thresh_from_dataset(
                value=value,
                unit=unit,
                threshold_var_name=threshold_var_name,
                reference_period=reference_period,
            )
        else:
            is_doy_per_threshold = unit == DOY_PERCENTILE_UNIT
        if isinstance(value, DataArray):
            self.prepare = None
            self._prepared_value = PercentileDataArray.from_da(value)
            self.is_ready = True
            self._initial_unit = None
            self._initial_value = None
            self.unit = self._prepared_value.attrs[UNITS_KEY]
        else:
            self.is_ready = False
            self._initial_unit = unit
            if not isinstance(value, list):
                value = [value]
            self._initial_value = [float(x) for x in value]
            self.unit = unit
        self.operator = operator
        self.threshold_var_name = threshold_var_name
        self.initial_query = initial_query
        self.threshold_min_value = threshold_min_value
        self.reference_period = reference_period
        self.doy_window_width = doy_window_width
        self.only_leap_years = only_leap_years
        self.interpolation = QuantileInterpolationRegistry.lookup(interpolation)
        self.unit = unit
        self.is_doy_per_threshold = is_doy_per_threshold

    def prepare(self, studied_data: DataArray) -> None:
        if self._initial_unit == DOY_PERCENTILE_UNIT:
            prepared_data = _build_doy_per(
                studied_data=studied_data,
                per_val=self._initial_value,
                reference_period=self.reference_period,
                interpolation=self.interpolation,
                only_leap_years=self.only_leap_years,
                doy_window_width=self.doy_window_width,
                percentile_min_value=self.threshold_min_value,
            )
        elif self._initial_unit == PERIOD_PERCENTILE_UNIT:
            prepared_data = _build_period_per(
                studied_data=studied_data,
                per_val=self._initial_value,
                reference_period=self.reference_period,
                interpolation=self.interpolation,
                only_leap_years=self.only_leap_years,
                percentile_min_value=self.threshold_min_value,
            )
        else:
            raise NotImplementedError(
                f"Unknown percentile unit" f" '{self._initial_unit}'."
            )
        self._prepared_value = prepared_data.chunk("auto")
        self.is_ready = True

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, PercentileThreshold)
            and self.initial_query == other.initial_query
            and self.doy_window_width == other.doy_window_width
            and self.only_leap_years == other.only_leap_years
            and self.interpolation == other.interpolation
            and self.reference_period == other.reference_period
            and self.unit == other.unit
            and self.threshold_min_value == other.threshold_min_value
        )

    # noinspection PyMethodOverriding
    # (reason: with * and **kwargs we can have a different signature while still
    # being liskov proof)
    def format_metadata(
        self,
        *,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
        src_freq: Frequency,
        must_run_bootstrap: bool = False,
        **kwargs,
    ) -> ThresholdMetadata:
        per_coord = self.value.coords["percentiles"]
        templates = self._get_metadata_templates(per_coord)
        climatology_bounds: list[str] = self.value.attrs.get("climatology_bounds")
        conf: PercentileTemplateConfig = {
            "climatology_bounds": climatology_bounds,
            "doy_window_width": self.doy_window_width,
            "src_freq": src_freq,
            "operator": self.operator,
            "unit": self.unit,
            "per_coord": per_coord.values,
            "threshold_min_value": self.threshold_min_value,
            "must_run_bootstrap": must_run_bootstrap,
        }
        conf.update(jinja_scope)
        return {
            k: jinja_env.from_string(v, globals=conf).render()
            for k, v in templates.items()
        }

    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,
    ) -> DataArray:
        if override_op is not None:
            op = override_op
        else:
            op = self.operator
        if self.is_ready:
            return self._per_compute(
                comparison_data,
                self.value,
                op,
                self.is_doy_per_threshold,
                kwargs.get("freq", None),
                kwargs.get("bootstrap", False),
            )
        else:
            raise RuntimeError(
                "This PercentileThreshold is not ready. You must first call `.prepare`"
                " with a `studied_data` parameter in order to prepare the threshold"
                " for computation."
            )

    def _get_metadata_templates(self, per_coord: DataArray) -> ThresholdMetadata:
        if self.is_doy_per_threshold:
            if per_coord.size == 1:
                return EN_THRESHOLD_TEMPLATE["single_doy_percentile"]
            else:
                return EN_THRESHOLD_TEMPLATE["multiple_doy_percentiles"]
        else:
            if per_coord.size == 1:
                return EN_THRESHOLD_TEMPLATE["single_period_percentile"]
            else:
                return EN_THRESHOLD_TEMPLATE["multiple_period_percentiles"]

    @percentile_bootstrap
    def _per_compute(
        self,
        comparison_data: xr.DataArray,
        per: xr.DataArray,
        op: Callable[[DataArray, DataArray], DataArray],
        is_doy_per_threshold: bool,
        freq: str,  # noqa used by @percentile_bootstrap
        bootstrap: bool,  # noqa used by @percentile_bootstrap
    ) -> DataArray:
        if is_doy_per_threshold:
            threshold_value = resample_doy(per, comparison_data)
        else:
            threshold_value = per
        return op(comparison_data, threshold_value)


class BasicThreshold(Threshold):
    """
    Pint ready simple threshold (e.g. "> 300 K").

    When built, ``value`` is always turned into a ``xarray.DataArray``.
    The ``unit`` property as a setter that will attempt a unit conversion using
    units found in xclim's pint registry.

    The actual unit can be overridden by modify ``value.attrs["units"]`` directly.
    """

    @property
    def unit(self) -> str | None:
        return self.value.attrs[UNITS_KEY]

    @unit.setter
    def unit(self, unit):
        if self.value.attrs.get(UNITS_KEY, None) is not None and unit is not None:
            self.value = convert_units_to(self.value, unit)
        self.value.attrs[UNITS_KEY] = unit

    def __init__(
        self,
        operator: Operator | str,
        value: ThresholdValueType,
        unit: str | None = None,
        initial_query: str | None = None,
        threshold_min_value: pint.Quantity | None = None,
        threshold_var_name: str | None = None,
        **kwargs,  # noqa
    ):
        if (
            is_number_sequence(value) or isinstance(value, (float, int))
        ) and threshold_min_value is not None:
            raise InvalidIcclimArgumentError(
                "Cannot use threshold_min_value with scalars"
            )
        if is_dataset_path(value) or isinstance(value, Dataset):
            # e.g. build_threshold(">", "thresh*.nc" , "degC")
            thresh_da = _get_dataarray_from_dataset(threshold_var_name, value)
            built_value = _apply_min_value(thresh_da, threshold_min_value)
            if unit is None:
                unit = built_value.attrs.get(UNITS_KEY, None)
        elif isinstance(value, DataArray):
            built_value = _apply_min_value(value, threshold_min_value)
            if unit is None:
                unit = built_value.attrs.get(UNITS_KEY, None)
        elif is_number_sequence(value):
            built_value = DataArray(
                name="threshold",
                data=value,
                attrs={UNITS_KEY: unit},
                dims="threshold",
                coords={"threshold": value},
            )
        elif isinstance(value, (float, int)):
            # e.g. build_threshold(">", [2,3,4], "degC")
            built_value = DataArray(
                name="threshold", data=value, attrs={UNITS_KEY: unit}
            )
        else:
            raise NotImplementedError(f"Cannot build threshold from a {type(value)}.")
        if unit is not None:
            built_value = convert_units_to(built_value, unit)
        self.operator = operator
        self.value = built_value
        self.unit = unit
        self.is_ready = True
        self.threshold_var_name = threshold_var_name
        self.initial_query = initial_query
        if unit is not None and (
            threshold_min_value is not None and threshold_min_value.dimensionless
        ):
            self.threshold_min_value = float(threshold_min_value.m) * str2pint(unit)
        else:
            self.threshold_min_value = threshold_min_value
        self.prepare = None

    def __eq__(self, other):
        return (
            isinstance(other, BasicThreshold)
            and self.operator == other.operator
            and self.value == other.value
            and self.unit == other.unit
            and self.initial_query == other.initial_query
            and self.threshold_min_value == other.threshold_min_value
        )

    def _get_metadata_templates(self) -> ThresholdMetadata:
        if self.value.size == 1:
            return EN_THRESHOLD_TEMPLATE["single_value"]
        else:
            return EN_THRESHOLD_TEMPLATE["multiple_values"]

    def format_metadata(
        self,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
        **kwargs,
    ) -> ThresholdMetadata:
        templates = self._get_metadata_templates()
        conf = {
            "operator": self.operator,
            "unit": self.unit,
            "threshold_min_value": self.threshold_min_value,
            "must_run_bootstrap": True,
            "value": self.value,
        }
        if self.value.size > 1:
            conf["min_value"] = np.format_float_positional(
                self.value.min().values[()], 3
            )
            conf["max_value"] = np.format_float_positional(
                self.value.max().values[()], 3
            )
        conf.update(jinja_scope)
        return {
            k: jinja_env.from_string(v, globals=conf).render()
            for k, v in templates.items()
        }

    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,
    ) -> DataArray:
        if override_op is not None:
            return override_op(comparison_data, self.value)
        return self.operator.compute(comparison_data, self.value)


def _build_period_per(
    studied_data: DataArray,
    per_val: Sequence[float],
    reference_period: Sequence[str],
    interpolation: QuantileInterpolation,
    only_leap_years: bool,
    percentile_min_value: pint.Quantity | None,
) -> PercentileDataArray:
    reference = build_reference_da(
        studied_data,
        reference_period,
        only_leap_years,
        percentile_min_value=percentile_min_value,
    )
    computed_per = xr.apply_ufunc(
        calc_perc,
        reference,
        input_core_dims=[["time"]],
        output_core_dims=[["percentiles"]],
        keep_attrs=True,
        kwargs=dict(
            percentiles=per_val,
            alpha=interpolation.alpha,
            beta=interpolation.beta,
            copy=True,
        ),
        dask="parallelized",
        output_dtypes=[reference.dtype],
        dask_gufunc_kwargs=dict(output_sizes={"percentiles": 1}, allow_rechunk=True),
    )
    computed_per = computed_per.assign_coords(
        percentiles=xr.DataArray(per_val, dims=("percentiles",))
    )
    res = PercentileDataArray.from_da(
        source=computed_per,
        climatology_bounds=build_climatology_bounds(reference),
    )
    return res


def _build_doy_per(
    studied_data: DataArray,
    per_val: Sequence[float],
    reference_period: Sequence[str],
    interpolation: QuantileInterpolation,
    only_leap_years: bool,
    doy_window_width: int,
    percentile_min_value: pint.Quantity | None,
) -> PercentileDataArray:
    reference = build_reference_da(
        studied_data,
        reference_period,
        only_leap_years,
        percentile_min_value,
    )
    res = percentile_doy(
        arr=reference,
        window=doy_window_width,
        per=per_val,
        alpha=interpolation.alpha,
        beta=interpolation.beta,
    ).compute()  # "optimization" (diminish dask scheduler workload)
    return res


def _read_string_threshold(query: str) -> tuple[str, str, float]:
    value = re.findall(r"-?\d+\.?\d*", query)
    if len(value) == 0:
        raise InvalidIcclimArgumentError(f"Cannot build threshold from '{query}'")
    value = value[0]
    value_index = query.find(value)
    operator = query[0:value_index].strip()
    if operator == "":
        operator = None
    if query.endswith(value):
        unit = None
    else:
        unit = query[value_index + len(value) :].strip()
    return operator, unit, float(value)


def _build_min_value(
    threshold_min_value: None | str | float | pint.Quantity,
    default_unit: str | None,
) -> pint.Quantity | None:
    if threshold_min_value is None:
        return None
    elif isinstance(threshold_min_value, xc_units.Quantity):
        return threshold_min_value
    elif isinstance(threshold_min_value, (float, int)):
        if (
            default_unit == PERIOD_PERCENTILE_UNIT
            or default_unit == DOY_PERCENTILE_UNIT
        ):
            unit = None
        else:
            unit = default_unit
        return xc_units.Quantity(value=threshold_min_value, units=unit)
    elif isinstance(threshold_min_value, str):
        operator, unit, value = _read_string_threshold(threshold_min_value)
        if operator is not None and operator != "" and operator != ">=":
            raise InvalidIcclimArgumentError(
                f"cannot compute threshold_min_value with"
                f" {operator}. You don't need to fill an"
                f" operator for this parameter."
            )
        return xc_units.Quantity(value=value, units=unit)
    else:
        raise NotImplementedError(
            f"Unknown type '{type(threshold_min_value)}' for `threshold_min_value`."
        )


def _read_input(
    query: str | None = None,
    operator: Operator | str = None,
    value: ThresholdValueType = None,
    unit: str | None = None,
    threshold_min_value: str | float | None = None,
    thresholds: tuple[Threshold, Threshold] | None = None,
    logical_link: str | LogicalLink | None = None,
    **kwargs,
) -> ThresholdBuilderInput:
    if _must_read_query(query, operator, value, unit):
        if _is_bounded_threshold_query(query):
            return _read_bounded_threshold_query(query)
        else:
            return _read_threshold_from_query(query, threshold_min_value, kwargs)
    elif _must_read_bounded(operator, value, unit, thresholds, logical_link):
        return _read_bounded_threshold(thresholds, logical_link)
    elif _must_read_from_args(operator, value):
        if (operator := OperatorRegistry.lookup(operator, no_error=True)) is None:
            operator = OperatorRegistry.REACH
        return {
            "operator": operator,
            "unit": unit,
            "value": value,
            "threshold_min_value": _build_min_value(threshold_min_value, unit),
            **kwargs,
        }
    else:
        raise NotImplementedError("Could not read threshold")


def _read_bounded_threshold(
    thresholds: tuple[Threshold, Threshold], logical_link: LogicalLink | str
) -> ThresholdBuilderInput:
    acc = []
    for t in thresholds:
        if isinstance(t, str):
            acc.append(_read_input(t))
        elif isinstance(t, Threshold):
            acc.append(t)
        else:
            raise NotImplementedError(f"Unknown type '{type(t)}'")
    if len(acc) > 2:
        raise NotImplementedError(
            "Can't build BoundedThreshold on more than 2 thresholds."
        )
    if isinstance(logical_link, str):
        logical_link = LogicalLinkRegistry.lookup(logical_link)
    return {  # noqa
        "initial_query": None,
        "thresholds": tuple(acc),
        "logical_link": logical_link,
    }


def _read_threshold_from_query(
    query: str,
    threshold_min_value: None | str | float | xc_units.Quantity,
    kwargs: dict,
) -> ThresholdBuilderInput:
    operator, unit, value = _read_string_threshold(query)
    if (operator := OperatorRegistry.lookup(operator, no_error=True)) is None:
        operator = OperatorRegistry.REACH
    return {
        "operator": operator,
        "unit": unit,
        "value": value,
        "threshold_min_value": _build_min_value(threshold_min_value, unit),
        "initial_query": query,
        **kwargs,
    }


def _must_read_query(
    query: str | None,
    operator: Operator | str,
    value: ThresholdValueType,
    unit: str | None,
) -> bool:
    return (
        isinstance(query, str) and operator is None and value is None and unit is None
    )


def _must_read_bounded(
    operator: Operator | str,
    value: ThresholdValueType,
    unit: str | None,
    thresholds: tuple[Threshold, Threshold] | None,
    logical_link: str | LogicalLink | None,
) -> bool:
    return (
        operator is None
        and value is None
        and unit is None
        and thresholds is not None
        and logical_link is not None
    )


def _must_read_from_args(operator: Operator | str, value: ThresholdValueType) -> bool:
    return operator is not None and value is not None


def _is_bounded_threshold_query(query: str) -> bool:
    return any(
        [l_l.name.upper() in query.upper() for l_l in LogicalLinkRegistry.values()]
    )


def _read_bounded_threshold_query(query: str) -> ThresholdBuilderInput:
    link = None
    split_word = None
    uppered = query.upper()
    for l_l in LogicalLinkRegistry.values():
        if l_l.name.upper() in uppered:
            link = l_l
            index_of_link = uppered.index(l_l.name.upper())
            split_word = query[index_of_link : index_of_link + len(l_l.name)]
            break
    if link is None:
        raise InvalidIcclimArgumentError(f"No logical link found in {query}")
    threshs = query.split(split_word)
    if len(threshs) != 2:
        raise InvalidIcclimArgumentError(
            "BoundedThreshold can only be built on 2"
            f" thresholds. We found {len(threshs)} here."
        )
    return {
        "initial_query": query,
        "thresholds": (_read_input(threshs[0]), _read_input(threshs[1])),
        "logical_link": link,
    }


def _must_build_per_threshold(input: ThresholdBuilderInput) -> bool:
    value = input.get("value")
    unit = input.get("unit", None)
    var_name = input.get("threshold_var_name", None)
    return _has_per_unit(unit, value) or _is_per_dataset(var_name, value)


def _is_per_dataset(threshold_var_name: str, value: str | Dataset | DataArray) -> bool:
    if is_dataset_path(value) or isinstance(value, Dataset):
        thresh_da = _get_dataarray_from_dataset(threshold_var_name, value)
    else:
        thresh_da = value
    return PercentileDataArray.is_compatible(thresh_da)


def _has_per_unit(unit: str | None | Sequence[float], value: float) -> bool:
    return isinstance(value, (float, Sequence)) and (
        unit == DOY_PERCENTILE_UNIT or unit == PERIOD_PERCENTILE_UNIT
    )


def _must_build_basic_threshold(input: ThresholdBuilderInput) -> bool:
    value = input.get("value")
    threshold_var_name = input.get("threshold_var_name", None)
    if is_dataset_path(value) or isinstance(value, Dataset):
        thresh_da = _get_dataarray_from_dataset(threshold_var_name, value)
        return not PercentileDataArray.is_compatible(thresh_da)
    return (
        is_number_sequence(value)
        or isinstance(value, (float, int))
        or isinstance(value, DataArray)
    )


def _must_build_bounded_threshold(input: ThresholdBuilderInput) -> bool:
    logical_link = input.get("logical_link")
    return logical_link is not None


def _apply_min_value(thresh_da: DataArray, min_value: pint.Quantity | None):
    if min_value is not None:
        if min_value.dimensionless:
            min_value = convert_units_to(min_value.m, thresh_da)
        else:
            min_value = convert_units_to(str(min_value), thresh_da)
        built_value = thresh_da.where(thresh_da > min_value, np.nan)
        return built_value
    else:
        return thresh_da


def _get_dataarray_from_dataset(
    threshold_var_name: str | None, value: Dataset | str
) -> DataArray:
    if isinstance(value, Dataset):
        ds = value
    else:
        ds = read_dataset(value, standard_var=None)
    if threshold_var_name is None:
        if len(ds.data_vars) == 1:
            threshold_var_name = get_name_of_first_var(ds)
        else:
            names = find_standard_vars(ds)
            if len(names) == 1:
                threshold_var_name = names[0]
            else:
                raise InvalidIcclimArgumentError(
                    f"Could not guess the variable to use as a threshold in {ds}."
                    f" Use `threshold_var_name` to specify which variable should be"
                    f" used."
                )
    thresh_da = ds[threshold_var_name]
    return thresh_da


def _build_per_thresh_from_dataset(
    value: ThresholdValueType,
    unit: str | None,
    threshold_var_name: str,
    reference_period: Sequence[datetime | str],
) -> tuple[DataArray, bool]:
    thresh_da = _get_dataarray_from_dataset(threshold_var_name, value)
    built_value = PercentileDataArray.from_da(
        standardize_percentile_dim_name(thresh_da),
        read_clim_bounds(reference_period, thresh_da),
    )
    if unit is not None:
        if built_value.attrs.get(UNITS_KEY, None) is not None:
            built_value = convert_units_to(built_value, unit)
        built_value.attrs[UNITS_KEY] = unit
    return built_value, DOY_COORDINATE in built_value.coords



from __future__ import annotations

import dataclasses
from typing import Any

from icclim.models.registry import Registry


@dataclasses.dataclass(init=False)
class IndexGroup:
    name: str
    values: list[IndexGroup]

    def __init__(self, name: str, values: list[IndexGroup] = None):
        self.name = name
        if values is None:
            self.values = [self]
        else:
            self.values = values

    def get_indices(self) -> list[Any]:
        # import locally to avoid circular dependency (an index has a IndexGroup)
        from icclim.ecad.ecad_indices import EcadIndexRegistry

        return list(
            filter(lambda i: i.group in self.values, EcadIndexRegistry.values())
        )

    def __or__(self, right):
        """Used to compose IndexGroup, e.g., IndexGroup1 | IndexGroup2."""
        if isinstance(right, IndexGroup):
            return IndexGroup(f"{self.name}_{right.name}", [self, right])
        raise NotImplementedError(
            f"Unexpected type for {right}: {type(right)}."
            f" An IndexGroup was expected."
        )

    def __eq__(self, other):
        if not isinstance(other, IndexGroup):
            return False
        return other.name == self.name


class IndexGroupRegistry(Registry[IndexGroup]):
    _item_class = IndexGroup

    TEMPERATURE = IndexGroup("temperature")
    HEAT = IndexGroup("heat")
    COLD = IndexGroup("cold")
    DROUGHT = IndexGroup("drought")
    RAIN = IndexGroup("rain")
    SNOW = IndexGroup("snow")
    WIND = IndexGroup("wind")
    # no climate index should be bounded to "all"
    WILD_CARD_GROUP = IndexGroup("all")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Sequence

import jinja2
import xarray
from xarray.core.dataarray import DataArray

from icclim.generic_indices.standard_variable import StandardVariable
from icclim.generic_indices.threshold import (
    PercentileThreshold,
    Threshold,
    build_threshold,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_types import InFileBaseType, InFileLike
from icclim.models.constants import REFERENCE_PERIOD_INDEX, UNITS_KEY
from icclim.models.frequency import Frequency, FrequencyRegistry
from icclim.models.global_metadata import GlobalMetadata
from icclim.models.standard_index import StandardIndex
from icclim.pre_processing.in_file_dictionary import InFileDictionary
from icclim.pre_processing.input_parsing import (
    DEFAULT_INPUT_FREQUENCY,
    build_reference_da,
    build_studied_data,
    guess_input_type,
    guess_var_names,
    read_dataset,
)


@dataclass
class ClimateVariable:
    """Internal icclim structure. It groups together the input variable (studied_data),
    its associated metadata (standard_variable) and the threshold it must be compared
    to.

    Attributes
    ----------
    name: str
        Name of the variable.
    standard_var: StandardVariable
        CF metadata bounded to the standard variable used for this ClimateVariable.
    studied_data: DataArray
        The variable studied.
    threshold: Threshold | None
        thresholds for this variable
    """

    name: str
    standard_var: StandardVariable | None
    studied_data: DataArray
    global_metadata: GlobalMetadata
    source_frequency: Frequency
    threshold: Threshold | None = None
    is_reference: bool = False

    def build_indicator_metadata(
        self,
        src_freq: Frequency,
        must_run_bootstrap: bool,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
    ) -> dict[str, str | dict]:
        metadata: dict[str, str | dict] = {"threshold": {}}
        if self.standard_var is None:
            metadata.update(
                dict(
                    standard_name="unknown_variable",
                    long_name="unknown variable",
                    short_name="input",
                )
            )
        else:
            metadata.update(self.standard_var.get_metadata())
        if self.threshold is not None:
            metadata.update(
                {
                    "threshold": self.threshold.format_metadata(
                        src_freq=src_freq,
                        must_run_bootstrap=must_run_bootstrap,
                        jinja_scope=jinja_scope,
                        jinja_env=jinja_env,
                    )
                }
            )
        return metadata


def build_climate_vars(
    climate_vars_dict: dict[str, InFileDictionary],
    ignore_Feb29th: bool,
    time_range: Sequence[str],
    base_period: Sequence[str] | None,
    standard_index: StandardIndex | None,
    is_compared_to_reference: bool,
) -> list[ClimateVariable]:
    if standard_index is not None and len(standard_index.input_variables) > len(
        climate_vars_dict
    ):
        raise InvalidIcclimArgumentError(
            f"Index {standard_index.short_name} needs"
            f" {len(standard_index.input_variables)} variables."
            f" Please provide them with an xarray.Dataset, netCDF file(s) or a"
            f" zarr store."
        )
    acc = []
    for i, raw_climate_var in enumerate(climate_vars_dict.items()):
        if standard_index is not None:
            standard_var = standard_index.input_variables[i]
        else:
            standard_var = None
        acc.append(
            _build_climate_var(
                raw_climate_var[0],
                raw_climate_var[1],
                ignore_Feb29th,
                time_range,
                standard_var=standard_var,
            )
        )
    if _standard_index_needs_ref(
        standard_index, is_compared_to_reference
    ) or _generic_index_needs_ref(standard_index, is_compared_to_reference):
        standard_var = (
            standard_index.input_variables[0] if standard_index is not None else None
        )
        added_var = _build_reference_variable(
            base_period,
            climate_vars_dict,
            standard_var=standard_var,
        )
        acc.append(added_var)
    return acc


def _standard_index_needs_ref(standard_index, is_compared_to_reference):
    return (
        standard_index
        and standard_index.qualifiers
        and REFERENCE_PERIOD_INDEX in standard_index.qualifiers
        and is_compared_to_reference
    )


def _generic_index_needs_ref(standard_index, is_compared_to_reference):
    return standard_index is None and is_compared_to_reference


def _build_reference_variable(
    reference_period: Sequence[str] | None,
    in_files,
    standard_var: StandardVariable,
) -> ClimateVariable:
    """This function add a secondary variable for indices such as anomaly that needs
    exactly two variables but where the second variable could just be a subset of the
    first one.
    """
    if reference_period is None:
        raise InvalidIcclimArgumentError(
            "Can't build a reference variable without a `base_period_time_range`"
        )
    var_name = list(in_files.keys())[0]
    if isinstance(in_files, dict):
        study_ds = read_dataset(
            list(in_files.values())[0]["study"],
            standard_var=standard_var,
            var_name=var_name,
        )
    else:
        study_ds = read_dataset(
            list(in_files.values())[0], standard_var=standard_var, var_name=var_name
        )
    studied_data = build_reference_da(
        study_ds[var_name],
        reference_period,
        only_leap_years=False,
        percentile_min_value=None,
    )
    return ClimateVariable(
        name=var_name + "_reference",
        standard_var=standard_var,
        studied_data=studied_data,
        threshold=None,
        global_metadata={
            "history": study_ds.attrs.get("history", None),
            "source": study_ds.attrs.get("source", None),
            "time_encoding": study_ds.time.encoding,
        },
        source_frequency=FrequencyRegistry.lookup(
            xarray.infer_freq(studied_data.time) or DEFAULT_INPUT_FREQUENCY
        ),
        is_reference=True,
    )


def read_in_files(
    in_files: InFileLike,
    var_names: Sequence[str] | None,
    threshold: Threshold | Sequence[Threshold] | None,
    standard_index: StandardIndex | None,
) -> dict[str, InFileDictionary]:
    if isinstance(in_files, dict):
        if var_names is not None:
            raise InvalidIcclimArgumentError(
                "`var_name` must be None when `in_files` is a dictionary."
                " The dictionary keys are used in place of `var_name`."
            )
        if isinstance(list(in_files.values())[0], dict):
            # case of in_files={tasmax: {"study": "tasmax.nc"}}
            return in_files
        else:
            # case of in_files={tasmax: "tasmax.nc"}
            return _build_in_file_dict(
                in_files=list(in_files.values()),
                standard_index=standard_index,
                threshold=threshold,
                var_names=list(in_files.keys()),
            )
    else:
        # case of in_files="tasmax.nc" and var_names="tasmax"
        return _build_in_file_dict(in_files, var_names, threshold, standard_index)


def _build_in_file_dict(
    in_files: InFileBaseType,
    var_names: Sequence[str],
    threshold: Threshold | Sequence[Threshold] | None,
    standard_index: StandardIndex | None,
):
    standard_var = (
        standard_index.input_variables[0] if standard_index is not None else None
    )
    input_dataset = read_dataset(
        in_files=in_files, standard_var=standard_var, var_name=var_names
    )
    var_names = guess_var_names(
        ds=input_dataset, standard_index=standard_index, var_names=var_names
    )
    if threshold is not None:
        if len(var_names) == 1:
            return {
                var_names[0]: {
                    "study": input_dataset[var_names[0]],
                    "thresholds": threshold,
                }
            }
        if not isinstance(threshold, Sequence):
            threshold = [threshold]
        if len(threshold) != len(var_names):
            # Allow 1 var with multiple thresholds or 1 threshold per var
            # but no other case
            raise InvalidIcclimArgumentError(
                "There must be as many thresholds as there are variables. There was"
                f" {len(threshold)} thresholds and {len(var_names)} variables."
            )
        return {
            var_name: {"study": input_dataset[var_name], "thresholds": threshold[i]}
            for i, var_name in enumerate(var_names)
        }
    else:
        return {var_name: {"study": input_dataset[var_name]} for var_name in var_names}


def _build_climate_var(
    climate_var_name: str,
    climate_var_data: InFileDictionary | InFileBaseType,
    ignore_Feb29th: bool,
    time_range: Sequence[str],
    standard_var: StandardVariable | None,
) -> ClimateVariable:
    if isinstance(climate_var_data, dict):
        study_ds = read_dataset(
            climate_var_data["study"], standard_var, climate_var_name
        )
        # todo: deprecate climate_var_data.get("per_var_name", None)
        #       for threshold_var_name
        climate_var_thresh = climate_var_data.get("thresholds", None)
    else:
        climate_var_data: InFileBaseType
        study_ds = read_dataset(climate_var_data, standard_var, climate_var_name)
        climate_var_thresh = None
    if standard_var is None:
        standard_var = guess_input_type(study_ds[climate_var_name])
    studied_data = build_studied_data(
        study_ds[climate_var_name],
        time_range,
        ignore_Feb29th,
        standard_var,
    )
    if climate_var_thresh is not None:
        climate_var_thresh = _build_threshold(
            climate_var_thresh=climate_var_thresh,
            original_data=study_ds[climate_var_name],
            conversion_unit=studied_data.attrs[UNITS_KEY],
        )
    return ClimateVariable(
        name=climate_var_name,
        standard_var=standard_var,
        studied_data=studied_data,
        threshold=climate_var_thresh,
        global_metadata={
            "history": study_ds.attrs.get("history", None),
            "source": study_ds.attrs.get("source", None),
            "time_encoding": study_ds.time.encoding,
        },
        source_frequency=FrequencyRegistry.lookup(
            xarray.infer_freq(studied_data.time) or DEFAULT_INPUT_FREQUENCY
        ),
    )


def _build_threshold(
    climate_var_thresh: str | Threshold,
    original_data: DataArray,
    conversion_unit: str,
) -> Threshold:
    if isinstance(climate_var_thresh, str):
        climate_var_thresh: Threshold = build_threshold(climate_var_thresh)
    if (
        isinstance(climate_var_thresh, PercentileThreshold)
        and not climate_var_thresh.is_ready
    ):
        climate_var_thresh.prepare(original_data)
    climate_var_thresh.unit = conversion_unit
    return climate_var_thresh



from __future__ import annotations

from typing import Any, TypedDict

# fmt: off
# flake8: noqa
from numpy import ndarray
from pint import Quantity


class ThresholdMetadata(TypedDict):
    long_name: str
    short_name: str
    standard_name: str

class ThresholdTemplateDict(TypedDict):
    single_value: ThresholdMetadata
    multiple_values: ThresholdMetadata
    single_doy_percentile: ThresholdMetadata
    multiple_doy_percentiles: ThresholdMetadata
    single_period_percentile: ThresholdMetadata
    multiple_period_percentiles: ThresholdMetadata
    bounded_threshold: ThresholdMetadata

class PercentileTemplateConfig(TypedDict, total=False):
    climatology_bounds: list[str]
    doy_window_width: int
    src_freq: Any
    operator: Any
    unit: str | None
    per_coord: ndarray | float
    threshold_min_value: Quantity | None
    must_run_bootstrap: bool

DOY_PER_METHOD = (
    "day of year percentiles were computed per grid cell, on the"
    " {{climatology_bounds}} period, with a"
    " {{doy_window_width}}"
    " {{src_freq.units}} centred window to aggregate values around each day of year")
PERIOD_PER_META = (
    "period percentiles were computed per grid cell, on the"
    " {{climatology_bounds}} period")
BOOTSTRAP_META = (
    "the bootstrap algorithm (Zhang, 2005) has been applied to compute doy"
    " percentiles for the period overlapping both the reference"
    " period and the studied period")
THRESHOLD_MIN_VALUE_TEMPLATE = (
    "(only values >= {{threshold_min_value}}"
    " {{unit}} were considered to compute thresholds)")

EN_THRESHOLD_TEMPLATE: ThresholdTemplateDict = {
    "single_value":   {
        "standard_name": "{{operator.standard_name}}_threshold",
        "long_name":     "{{operator.long_name}}"
                         " {{value.values[()]}}"
                         " {{unit}}",
        "short_name":    "{{operator.short_name}}_threshold",
    },
    "multiple_values": {
        "standard_name": "{{operator.standard_name}}_thresholds",
        "long_name":     "{{operator.long_name}}"
                         "{% if value.size < 10 %}"
                             " {{value.values}}"
                             " {{unit}}"
                         "{% else %}"
                             " per grid cell values between"
                             " {{min_value}}"
                             " {{unit}}"
                             " and {{max_value}}"
                             " {{unit}}."
                         "{% endif %}"
                         "{% if threshold_min_value %}"
                             f" {THRESHOLD_MIN_VALUE_TEMPLATE}"
                         "{% endif %}",
        "short_name":    "{{operator.short_name}}_threshold",
    },
    "single_doy_percentile": {
        "standard_name": "{{operator.standard_name}}_doy_percentile_threshold",
        "long_name":     "{{operator.long_name}}"
                         " {{per_coord}}th day of year percentile"
                         "{% if threshold_min_value %}"
                             f" {THRESHOLD_MIN_VALUE_TEMPLATE}"
                         "{% endif %}",
        "short_name":    "doy_per_threshold",
    },
    "multiple_doy_percentiles": {
        "standard_name": "{{operator.standard_name}}_doy_percentile_thresholds",
        "long_name":     "{{operator.long_name}}"
                         " {{per_coord}} day of year percentiles"
                         "{% if threshold_min_value %}"
                             f" {THRESHOLD_MIN_VALUE_TEMPLATE}"
                         "{% endif %}",
        "short_name":    "doy_per_thresholds",
    },
    "single_period_percentile": {
        "standard_name": "{{operator.standard_name}}_period_percentile_threshold",
        "long_name":     "{{operator.long_name}}"
                         " {{per_coord}}th period percentile"
                         "{% if threshold_min_value %}"
                             f" {THRESHOLD_MIN_VALUE_TEMPLATE}"
                         "{% endif %}",
        "short_name":    "period_per_threshold",
    },
    "multiple_period_percentiles": {
        "standard_name": "{{operator.standard_name}}_period_percentile_thresholds",
        "long_name":     "{{operator.long_name}}"
                         " {{per_coord}} period percentiles"
                         "{% if threshold_min_value %}"
                             f" {THRESHOLD_MIN_VALUE_TEMPLATE}"
                         "{% endif %}",
        "short_name":    "period_per_thresholds",
    },
    "bounded_threshold": {
        "standard_name": "{{left_threshold.standard_name}}"
                         "_{{logical_link.standard_name}}"
                         "_{{right_threshold.standard_name}}",
        "long_name":     "{{left_threshold.long_name}}"
                         " {{logical_link.long_name}}"
                         " {{right_threshold.long_name}}",
        "short_name":    "{{left_threshold.short_name}}"
                         "_{{logical_link.short_name}}"
                         "_{{right_threshold.short_name}}",
    },
}



from __future__ import annotations

import pytest

from icclim.ecad.ecad_indices import EcadIndexRegistry
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.standard_index import StandardIndex


def test_listing():
    res = EcadIndexRegistry.list()
    indices = [
        k for k, v in EcadIndexRegistry.__dict__.items() if isinstance(v, StandardIndex)
    ]
    assert len(res) == len(indices)


class Test_index_from_string:
    def test_simple(self):
        res = EcadIndexRegistry.lookup("SU")
        assert res == EcadIndexRegistry.SU

    def test_lowercase(self):
        res = EcadIndexRegistry.lookup("tx90p")
        assert res == EcadIndexRegistry.TX90P

    def test_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            EcadIndexRegistry.lookup("cacahuête")



from __future__ import annotations

import numpy as np
import pandas as pd
import xarray
import xarray as xr

from icclim.models.constants import UNITS_KEY

VALUE_COUNT = 365 * 5 + 1  # 5 years of data (with 1 leap year)
COORDS = dict(
    lat=[42],
    lon=[42],
    time=pd.date_range("2042-01-01", periods=VALUE_COUNT, freq="D"),
)
K2C = 273.15

CF_TIME_RANGE = xr.cftime_range("2042-01-01", periods=VALUE_COUNT, freq="D")


def stub_tas(tas_value: float = 1.0, use_dask=False, use_cftime=False):
    da = xarray.DataArray(
        data=(np.full(VALUE_COUNT, tas_value).reshape((VALUE_COUNT, 1, 1))),
        dims=["time", "lat", "lon"],
        coords=COORDS,
        attrs={UNITS_KEY: "K"},
    )
    if use_cftime:
        da["time"] = CF_TIME_RANGE
    if use_dask:
        da.chunk()
    return da


def stub_pr(value: float, use_dask=False):
    da = xarray.DataArray(
        data=np.full(VALUE_COUNT, value).reshape((VALUE_COUNT, 1, 1)),
        coords=COORDS,
        dims=["time", "lat", "lon"],
        name="pr",
        attrs={UNITS_KEY: "kg m-2 d-1"},
    )
    if use_dask:
        da.chunk()
    return da



from __future__ import annotations

import os
from typing import Callable

import numpy as np
import pandas as pd
import pint
import pytest
import xarray as xr
from xclim.core.calendar import percentile_doy
from xclim.core.units import units as xc_units

from icclim.generic_indices.threshold import (
    BasicThreshold,
    BoundedThreshold,
    PercentileThreshold,
    build_threshold,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import UNITS_KEY
from icclim.models.logical_link import LogicalLinkRegistry
from icclim.models.operator import OperatorRegistry


def test_value_error():
    with pytest.raises(NotImplementedError):
        build_threshold(value={"random": "stuff"})


def test_query_error():
    with pytest.raises(InvalidIcclimArgumentError):
        build_threshold("Coco l'asticot")


def test_build_threshold__from_query():
    res = build_threshold(">10degC")
    assert isinstance(res, BasicThreshold)
    assert res.operator == OperatorRegistry.GREATER
    assert res.value == 10
    assert res.unit == "degC"


def test_build_bounded_threshold__from_query():
    res = build_threshold(">10degC and <20degC")
    assert isinstance(res, BoundedThreshold)
    assert res.left_threshold.operator == OperatorRegistry.GREATER
    assert res.left_threshold.value == 10
    assert res.left_threshold.unit == "degC"
    assert res.logical_link == LogicalLinkRegistry.LOGICAL_AND
    assert res.right_threshold.operator == OperatorRegistry.LOWER
    assert res.right_threshold.value == 20
    assert res.right_threshold.unit == "degC"


def test_build_bounded_threshold__unit_conversion():
    res = build_threshold(">10degC and <300 K")
    res.unit = "degree_Fahrenheit"
    np.testing.assert_almost_equal(res.left_threshold.value, 50)
    np.testing.assert_almost_equal(res.right_threshold.value, 80.33)
    assert res.left_threshold.unit == "degree_Fahrenheit"
    assert res.right_threshold.unit == "degree_Fahrenheit"


def test_build_bounded_threshold__unit_conversion_erorr():
    # GIVEN
    res = build_threshold(">10degC and <300 K")
    # THEN
    with pytest.raises(pint.DimensionalityError):
        # WHEN
        res.unit = "meter"


def test_build_bounded_threshold__error():
    with pytest.raises(NotImplementedError):
        build_threshold(thresholds=[">10degC", ">11degC", ">12degC"], logical_link="or")


def test_build_bounded_threshold__from_and():
    t1 = build_threshold(">10degC")
    t2 = build_threshold(">12 doy_per")
    t3 = t1 & t2
    assert isinstance(t3, BoundedThreshold)
    assert isinstance(t3.left_threshold, BasicThreshold)
    assert t3.left_threshold.operator == OperatorRegistry.GREATER
    assert t3.left_threshold.value == 10
    assert t3.left_threshold.unit == "degC"
    assert t3.logical_link == LogicalLinkRegistry.LOGICAL_AND
    assert isinstance(t3.right_threshold, PercentileThreshold)
    assert t3.right_threshold.is_ready is False
    assert t3.right_threshold.operator == OperatorRegistry.GREATER
    assert t3.right_threshold._initial_unit == "doy_per"
    assert t3.right_threshold._initial_value == [12]


def test_build_bounded_threshold__from_or():
    t1 = build_threshold(">10degC")
    t2 = build_threshold(">12 doy_per")
    t3 = t1 | t2
    assert isinstance(t3, BoundedThreshold)
    assert isinstance(t3.left_threshold, BasicThreshold)
    assert t3.left_threshold.operator == OperatorRegistry.GREATER
    assert t3.left_threshold.value == 10
    assert t3.left_threshold.unit == "degC"
    assert t3.logical_link == LogicalLinkRegistry.LOGICAL_OR
    assert isinstance(t3.right_threshold, PercentileThreshold)
    assert t3.right_threshold.is_ready is False
    assert t3.right_threshold.operator == OperatorRegistry.GREATER
    assert t3.right_threshold._initial_unit == "doy_per"
    assert t3.right_threshold._initial_value == [12]


def test_build_bounded_threshold__from_args():
    t1 = build_threshold(">10degC")
    t2 = build_threshold(">12 doy_per")
    t3 = build_threshold(
        thresholds=(t1, t2), logical_link=LogicalLinkRegistry.LOGICAL_OR
    )
    assert isinstance(t3, BoundedThreshold)
    assert isinstance(t3.left_threshold, BasicThreshold)
    assert t3.left_threshold.operator == OperatorRegistry.GREATER
    assert t3.left_threshold.value == 10
    assert t3.left_threshold.unit == "degC"
    assert t3.logical_link == LogicalLinkRegistry.LOGICAL_OR
    assert isinstance(t3.right_threshold, PercentileThreshold)
    assert t3.right_threshold.is_ready is False
    assert t3.right_threshold.operator == OperatorRegistry.GREATER
    assert t3.right_threshold._initial_unit == "doy_per"
    assert t3.right_threshold._initial_value == [12]


def test_basic_threshold_eq():
    a = build_threshold(">10degC")
    b = build_threshold(">10degC")
    c = build_threshold(">10mm")
    assert a == b
    assert a != c


def test_percentile_threshold_eq():
    a = build_threshold(">10doy_per")
    b = build_threshold(">10doy_per")
    c = build_threshold(">20doy_per")
    assert a == b
    assert a != c


def test_bounded_threshold_eq():
    a = build_threshold(">10doy_per")
    a_bis = build_threshold(">10doy_per")
    b = build_threshold(">15doy_per")
    c = build_threshold(">20doy_per")
    assert a & b == a_bis & b
    assert a & b == b & a
    assert a & b != a & c


def test_per_threshold_min_value__operand_error():
    with pytest.raises(InvalidIcclimArgumentError):
        build_threshold(">10doy_per", threshold_min_value="< 10 degC")


def test_per_threshold_min_value__type_error():
    with pytest.raises(NotImplementedError):
        build_threshold(">10doy_per", threshold_min_value=dict(random="stuff"))  # noqa


def test_per_threshold_min_value__string():
    a = build_threshold(">10doy_per", threshold_min_value="10 degC")
    assert a.threshold_min_value == xc_units.Quantity(10, "degC")


def test_per_threshold_min_value__quantity():
    a = build_threshold(">10doy_per", threshold_min_value=xc_units.Quantity(10, "degC"))
    assert a.threshold_min_value == xc_units.Quantity(10, "degC")


def test_per_threshold_min_value__number():
    a = build_threshold(">10doy_per", threshold_min_value=10)
    assert a.threshold_min_value.dimensionless
    assert a.threshold_min_value.magnitude == 10


def test_threshold_min_value__number():
    with pytest.raises(InvalidIcclimArgumentError):
        build_threshold(">10degC", threshold_min_value=5)


def test_threshold_min_value__error():
    with pytest.raises(NotImplementedError):
        build_threshold(None)


def test_build_per_threshold__from_query():
    res = build_threshold("<= 99 doy_per")
    assert isinstance(res, PercentileThreshold)
    assert res.operator == OperatorRegistry.LOWER_OR_EQUAL
    assert res._initial_value == [99]
    assert res.unit == "doy_per"  # not computed yet
    assert res._initial_unit == "doy_per"
    assert res.is_ready is False
    assert isinstance(res.prepare, Callable)
    with pytest.raises(RuntimeError):  # not computed yet
        res.value  # noqa


def test_build_basic_threshold__from_dataarray():
    TIME_RANGE = pd.date_range(start="2042-01-01", end="2045-12-31", freq="D")
    data = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(lat=[42], lon=[42], time=TIME_RANGE),
        attrs={UNITS_KEY: "degC"},
        name="toto",
    )
    res = build_threshold(operator=">=", value=data, threshold_min_value="280K")
    assert isinstance(res, BasicThreshold)
    assert res.operator == OperatorRegistry.GREATER_OR_EQUAL
    xr.testing.assert_equal(res.value, data)
    assert res.unit == "degC"
    assert res.is_ready is True


def test_build_basic_threshold__from_dataset():
    TIME_RANGE = pd.date_range(start="2042-01-01", end="2045-12-31", freq="D")
    ds = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(lat=[42], lon=[42], time=TIME_RANGE),
        attrs={UNITS_KEY: "degC"},
        name="tas",
    ).to_dataset()
    ds["tutu"] = ds["tas"]
    res = build_threshold(operator=">=", value=ds, threshold_min_value="280K")
    assert isinstance(res, BasicThreshold)
    assert res.operator == OperatorRegistry.GREATER_OR_EQUAL
    xr.testing.assert_equal(res.value, ds["tas"])
    assert res.unit == "degC"
    assert res.is_ready is True


def test_build_basic_threshold__from_dataset__error():
    TIME_RANGE = pd.date_range(start="2042-01-01", end="2045-12-31", freq="D")
    ds = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(lat=[42], lon=[42], time=TIME_RANGE),
        attrs={UNITS_KEY: "degC"},
        name="toto",
    ).to_dataset()
    ds["tutu"] = ds.toto
    with pytest.raises(InvalidIcclimArgumentError):
        # multiple variable without any recognizable one
        build_threshold(operator=">=", value=ds, threshold_min_value="280K")


class Test_FileBased:
    IN_FILE_PATH = "in.nc"
    TIME_RANGE = pd.date_range(start="2042-01-01", end="2045-12-31", freq="D")

    data = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(lat=[42], lon=[42], time=TIME_RANGE),
        attrs={UNITS_KEY: "degC"},
        name="toto",
    )

    @pytest.fixture(autouse=True)
    def cleanup(self):
        # setup
        # ...
        yield
        # teardown
        try:
            os.remove(self.IN_FILE_PATH)
        except FileNotFoundError:
            pass

    def test_build_basic_threshold__from_file(self):
        self.data.to_netcdf(path=self.IN_FILE_PATH)
        res = build_threshold(operator=">=", value=self.IN_FILE_PATH)
        assert isinstance(res, BasicThreshold)
        assert res.operator == OperatorRegistry.GREATER_OR_EQUAL
        xr.testing.assert_equal(res.value, self.data)
        assert res.unit == "degC"
        assert res.is_ready is True

    def test_threshold_min_value__number_from_file(self):
        self.data.to_netcdf(path=self.IN_FILE_PATH)
        res = build_threshold(
            operator=">=", value=self.IN_FILE_PATH, threshold_min_value=5
        )
        assert res.threshold_min_value == xc_units.Quantity(5, "degC")

    def test_build_percentile_threshold__from_file(self):
        doys = percentile_doy(self.data)
        doys.to_netcdf(path=self.IN_FILE_PATH)
        res = build_threshold(operator=">=", value=self.IN_FILE_PATH)
        assert isinstance(res, PercentileThreshold)
        assert res.operator == OperatorRegistry.GREATER_OR_EQUAL
        xr.testing.assert_equal(res.value, doys)
        assert res._initial_unit is None
        assert res.unit is None
        assert res.is_ready is True
        assert res._initial_value is None
        assert res.prepare is None



from __future__ import annotations

import datetime
from typing import Literal, Sequence, TypedDict

from icclim.models.logical_link import LogicalLink
from icclim.user_indices.calc_operation import CalcOperation, CalcOperationLike


class UserIndexDict(TypedDict, total=False):
    index_name: str
    calc_operation: CalcOperationLike | CalcOperation
    logical_operation: str | None | Sequence[str]  # >= | <= | ...| ==
    thresh: str | float | int | Sequence[str] | Sequence[float] | Sequence[int] | None
    extreme_mode: Literal["min", "max"] | None

    link_logical_operations: Literal["and", "or"] | LogicalLink | None
    coef: float | None
    date_event: bool
    var_type: Literal["t", "p"] | None
    window_width: int | None
    ref_time_range: list[datetime] | list[str] | tuple[str, str] | None

    # -- deprecated
    indice_name: str | None



from __future__ import annotations

from datetime import datetime
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

import icclim
from icclim.ecad.ecad_indices import EcadIndexRegistry
from icclim.generic_indices.generic_indicators import GenericIndicatorRegistry
from icclim.generic_indices.threshold import build_threshold
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_logger import VerbosityRegistry
from icclim.models.constants import QUANTILE_BASED, REFERENCE_PERIOD_INDEX
from icclim.models.frequency import FrequencyRegistry
from icclim.models.netcdf_version import NetcdfVersionRegistry
from icclim.models.quantile_interpolation import QuantileInterpolationRegistry
from icclim.models.standard_index import StandardIndex
from icclim.tests.testing_utils import stub_tas
from icclim.user_indices.calc_operation import CalcOperation, CalcOperationRegistry

DEFAULT_ARGS = dict(
    in_files="pouet.nc",
    var_name=None,
    slice_mode=FrequencyRegistry.YEAR,
    time_range=None,
    out_file=None,
    ignore_Feb29th=False,
    netcdf_version=NetcdfVersionRegistry.NETCDF4,
    logs_verbosity=VerbosityRegistry.LOW,
    date_event=False,
)


def build_expected_args(index: StandardIndex):
    expected_call_args = {"index_name": index.short_name.upper()}
    expected_call_args.update(DEFAULT_ARGS)
    qualifiers = [] if index.qualifiers is None else index.qualifiers
    if QUANTILE_BASED in qualifiers:
        expected_call_args.update(
            {
                "base_period_time_range": None,
                "only_leap_years": False,
                "interpolation": QuantileInterpolationRegistry.MEDIAN_UNBIASED.name,
                "save_thresholds": False,
            }
        )
    elif REFERENCE_PERIOD_INDEX in qualifiers:
        expected_call_args.update(
            {
                "base_period_time_range": None,
            }
        )
    if index.threshold is not None:
        if isinstance(index.threshold, str):
            t = build_threshold(index.threshold)
        elif isinstance(index.threshold, (list, tuple)):
            t = []
            for thresh in index.threshold:
                if isinstance(thresh, str):
                    t.append(build_threshold(thresh))
                else:
                    t.append(thresh)
        else:
            t = index.threshold
        expected_call_args.update({"threshold": t})
    expected_call_args.update({"out_unit": index.output_unit})
    return expected_call_args


@patch("icclim.index")
def test_generated_api(generic_index_fun_mock: MagicMock):
    for i in EcadIndexRegistry.values():
        # print(i)
        # GIVEN
        api_index_fun = eval(f"icclim.{i.short_name.lower()}")
        # WHEN
        api_index_fun(**DEFAULT_ARGS)
        # THEN
        expected_call_args = build_expected_args(i)
        generic_index_fun_mock.assert_called_with(**expected_call_args)
    for g in GenericIndicatorRegistry.values():
        print(g)
        # GIVEN
        api_index_fun = eval(f"icclim.{g.name.lower()}")
        # WHEN
        api_index_fun(**DEFAULT_ARGS)
        generic_index_fun_mock.assert_called()


@patch("icclim.index")
def test_custom_index(index_fun_mock: MagicMock):
    user_index_args = dict(
        in_files="pouet_file.nc",
        var_name=None,
        slice_mode=FrequencyRegistry.YEAR,
        time_range=None,
        out_file=None,
        base_period_time_range=None,
        only_leap_years=False,
        ignore_Feb29th=False,
        out_unit=None,
        netcdf_version=NetcdfVersionRegistry.NETCDF4,
        logs_verbosity=VerbosityRegistry.LOW,
        doy_window_width=5,
        save_thresholds=False,
        date_event=False,
        sampling_method="resample",
        min_spell_length=6,
        rolling_window_width=5,
        interpolation="median_unbiased",
        user_index={
            "index_name": "pouet",
            "calc_operation": "nb_events",
            "logical_operation": "gt",
            "thresh": 0,
            "date_event": True,
        },
    )
    icclim.custom_index(**user_index_args)
    index_fun_mock.assert_called_with(**user_index_args)


# integration test
def test_txx__season_slice_mode():
    # GIVEN
    tas = stub_tas()
    tas.loc[{"time": "2043-02-02"}] = 295
    tas.loc[{"time": "2043-01-01"}] = 303.15  # 30°C 273.15
    # WHEN
    res = icclim.txx(tas, slice_mode=["season", [11, 12, 1, 2]]).compute()
    # THEN
    # missing values for nov, dec of first period
    np.testing.assert_array_equal(res.TXx.isel(time=0), np.NAN)
    np.testing.assert_array_equal(res.TXx.isel(time=1), 30.0)
    np.testing.assert_array_equal(
        res.time_bounds.isel(time=0),
        [np.datetime64("2041-11-01"), np.datetime64("2042-02-28")],
    )


def test_txx__months_slice_mode():
    tas = stub_tas()
    tas.loc[{"time": "2042-11-02"}] = 295
    tas.loc[{"time": "2042-01-01"}] = 303.15  # 30°C 273.15
    res = icclim.txx(tas, slice_mode=["months", [11, 1]]).compute()
    np.testing.assert_array_equal(res.TXx.isel(time=0), 30)
    np.testing.assert_array_equal(res.TXx.isel(time=1), np.NAN)
    np.testing.assert_almost_equal(res.TXx.sel(time="2042-11"), 21.85)
    np.testing.assert_array_equal(
        res.time_bounds.isel(time=0),
        [np.datetime64("2042-01-01"), np.datetime64("2042-01-31")],
    )


# integration test
@pytest.mark.parametrize(
    "operator, expectation_year_1, expectation_year_2",
    [
        (CalcOperationRegistry.MIN, 303.15, 280.15),
        (CalcOperationRegistry.MAX, 303.15, 280.15),
        (CalcOperationRegistry.SUM, 303.15, 280.15),
        # values below 275 are filtered out
        (CalcOperationRegistry.MEAN, 303.15, 280.15),
        (CalcOperationRegistry.EVENT_COUNT, 1, 1),
        (CalcOperationRegistry.MAX_NUMBER_OF_CONSECUTIVE_EVENTS, 1, 1),
    ],
)
def test_custom_index__season_slice_mode(
    operator: CalcOperation, expectation_year_1, expectation_year_2
):
    tas = stub_tas(275.0)
    tas.loc[{"time": "2043-01-01"}] = 303.15
    tas.loc[{"time": "2043-12-01"}] = 280.15
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        var_name="a_name",
        user_index={
            "index_name": "pouet",
            "calc_operation": operator,
            "logical_operation": "gt",
            "thresh": 275,
        },
    ).compute()
    # missing values algo applied for first and last years
    np.testing.assert_almost_equal(res.pouet.isel(time=0), np.NAN)
    np.testing.assert_almost_equal(res.pouet.isel(time=-1), np.NAN)
    np.testing.assert_almost_equal(res.pouet.isel(time=1), expectation_year_1)
    np.testing.assert_almost_equal(res.pouet.isel(time=2), expectation_year_2)


# integration test
@pytest.mark.parametrize(
    "operator, expectation_year_1, expectation_year_2",
    [
        (CalcOperationRegistry.RUN_MEAN, 275.0, 276.0),
        (CalcOperationRegistry.RUN_SUM, 1925.0, 1932.0),
    ],
)
def test_custom_index_run_algos__season_slice_mode(
    operator, expectation_year_1, expectation_year_2
):
    tas = stub_tas(275.0)
    tas.loc[{"time": "2043-12-01"}] = 282.0
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        var_name="a_name",
        user_index={
            "index_name": "pouet",
            "calc_operation": operator,
            "extreme_mode": "max",
            "window_width": 7,
        },
    )
    # missing values algo applied for first and last years
    np.testing.assert_almost_equal(res.pouet.isel(time=0), np.NAN)
    np.testing.assert_almost_equal(res.pouet.isel(time=-1), np.NAN)
    np.testing.assert_almost_equal(res.pouet.isel(time=1), expectation_year_1)
    np.testing.assert_almost_equal(res.pouet.isel(time=2), expectation_year_2)


def test_custom_index_anomaly__error_single_var():
    tas = stub_tas(2.0)
    with pytest.raises(InvalidIcclimArgumentError):
        # error: it needs 2 vars or 1 var and a ref period
        icclim.custom_index(
            in_files=tas,
            user_index={
                "index_name": "anomaly",
                "calc_operation": CalcOperationRegistry.ANOMALY,
            },
        )


def test_custom_index_anomaly__error_():
    tas = stub_tas(2.0)
    with pytest.raises(InvalidIcclimArgumentError):
        # error: Can't resample the reference variable if it is already a
        # subsample of the studied variable. (need another sampling_method)
        icclim.custom_index(
            in_files=tas,
            slice_mode=["season", [12, 1]],
            base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
            user_index={
                "index_name": "anomaly",
                "calc_operation": CalcOperationRegistry.ANOMALY,
            },
        )


def test_custom_index_anomaly__datetime_ref_period():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
        sampling_method="groupby_ref_and_resample_study",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperationRegistry.ANOMALY,
        },
    ).compute()
    # missing values algo applied for first and last years
    np.testing.assert_almost_equal(res.anomaly.sel(time="2041"), np.NAN)
    np.testing.assert_almost_equal(res.anomaly.sel(time="2042"), 0)
    np.testing.assert_almost_equal(res.anomaly.sel(time="2043"), 0)
    np.testing.assert_almost_equal(res.anomaly.sel(time="2044"), 4.80645161)
    np.testing.assert_almost_equal(res.anomaly.sel(time="2045"), 0)
    np.testing.assert_almost_equal(res.anomaly.sel(time="2046"), np.NAN)


def test_custom_index_anomaly__groupby_and_resample_month():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode="month",
        base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
        sampling_method="groupby_ref_and_resample_study",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperationRegistry.ANOMALY,
        },
    ).compute()
    np.testing.assert_almost_equal(res.anomaly.sel(time="2045-01"), 9.61290323)


def test_custom_index_anomaly__groupby_and_resample_year():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode="year",
        base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
        sampling_method="groupby_ref_and_resample_study",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperationRegistry.ANOMALY,
        },
    ).compute()
    np.testing.assert_almost_equal(res.anomaly.sel(time="2045"), 0.81643836)


def test_custom_index_anomaly__groupby_and_resample_day():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode="day",
        base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
        sampling_method="groupby_ref_and_resample_study",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperationRegistry.ANOMALY,
        },
    ).compute()
    np.testing.assert_almost_equal(res.anomaly.sel(time="2045-01-01"), 298)


def test_custom_index_anomaly__groupby_and_resample_hour():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    with pytest.raises(NotImplementedError):
        icclim.custom_index(
            in_files=tas,
            slice_mode="hour",
            base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
            sampling_method="groupby_ref_and_resample_study",
            user_index={
                "index_name": "anomaly",
                "calc_operation": CalcOperationRegistry.ANOMALY,
            },
        )


def test_custom_index_anomaly__grouby_season():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
        sampling_method="groupby",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperationRegistry.ANOMALY,
        },
    ).compute()
    # missing values algo applied for first and last years
    np.testing.assert_almost_equal(res.anomaly, 0.96129032)


def test_custom_index_anomaly__grouby_month():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode="month",
        base_period_time_range=[datetime(2042, 1, 1), datetime(2044, 12, 31)],
        sampling_method="groupby",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperationRegistry.ANOMALY,
        },
    ).compute()
    # missing values algo applied for first and last years
    assert len(res.anomaly.month) == 12
    np.testing.assert_almost_equal(res.anomaly.sel(month=2), 0)
    np.testing.assert_almost_equal(res.anomaly.sel(month=1), 1.92258065)



from __future__ import annotations

# fmt: off
# flake8: noqa

ICCLIM_VERSION = "6.1.1"

# placeholders for user_index
PERCENTILE_THRESHOLD_STAMP = "p"
WET_DAY_THRESHOLD = 1  # 1mm
USER_INDEX_PRECIPITATION_STAMP = "p"
USER_INDEX_TEMPERATURE_STAMP = "t"

# percentiles dimension from percentile_doy
PERCENTILES_COORD = "percentiles"
# attribut holding the in_base time bounds
REFERENCE_PERIOD_ID = "reference_epoch"
# coordinate of day of year values (usually from 1 to 365/366)
DOY_COORDINATE = "dayofyear"
# Units attribute key for DataArray(s)
UNITS_KEY = "units"


# Aliases of input percentiles variables names
# Source icclim dev
VALID_PERCENTILE_DIMENSION = ["quantile", "percentile", "per", "centile"]

# Source of ECA&D indices definition
ECAD_ATBD = "ECA&D, Algorithm Theoretical Basis Document (ATBD) v11"

# Index qualifiers (needed to generate the API)
QUANTILE_BASED = "QUANTILE_BASED"
REFERENCE_PERIOD_INDEX = "REFERENCE_PERIOD_INDEX"


# Map of months index to their short name, used to get a pandas frequency anchor
MONTHS_MAP = {1:"JAN",  2:"FEB", 3:"MAR", 4:"APR", 5:"MAY", 6:"JUN", 7:"JUL", 8:"AUG", 9:"SEP", 10:"OCT", 11:"NOV", 12:"DEC" }

# Season defined by their month numbers
AMJJAS_MONTHS:list[int] = [*range(4, 9)]
ONDJFM_MONTHS:list[int] = [10, 11, 12, 1, 2, 3]
DJF_MONTHS:list[int] = [12, 1, 2]
MAM_MONTHS:list[int] = [*range(3, 6)]
JJA_MONTHS:list[int] = [*range(6, 9)]
SON_MONTHS:list[int] = [*range(9, 12)]

# pseudo units used with Threshold class (not in Pint)
PERIOD_PERCENTILE_UNIT = "period_per"
DOY_PERCENTILE_UNIT = "doy_per"

# Mapping of frequencies to generate metadata
# copied from xclim and updated.
# todo: we should probably turn that mapping to jinja templates to make it easier to maintain
EN_FREQ_MAPPING = {
    "YS": "year(s)", "Y": "year(s)", "AS": "year(s)", "A": "year(s)",
    "QS": "season(s)", "Q": "season(s)",
    "MS": "month(s)", "M": "month(s)",
    "W": "week(s)",
    "D": "day(s)",
    "H": "hour(s)",
    "JAN": "January starting", "FEB": "February starting", "MAR": "March starting", "APR": "April starting", "MAY": "May starting", "JUN": "June starting", "JUL": "July starting", "AUG": "August starting", "SEP": "September starting", "OCT": "October starting", "NOV": "November starting", "DEC": "December starting",
    # Arguments to "indexer"
    "DJF": "wintry", "MAM": "springlong", "JJA": "summery", "SON": "autumnal",
    "norm": "Normal",
    "m1": "january",  "m2": "february",  "m3": "march",  "m4": "april",  "m5": "may",  "m6": "june",  "m7": "july",  "m8": "august",  "m9": "september",  "m10": "october",  "m11": "november",  "m12": "december",
    "MON": "monday starting", "TUE": "tuesday starting", "WED": "wednesday starting", "THU": "thursday starting", "FRI": "friday starting", "SAT": "saturday starting", "SUN": "sunday starting"
}
FREQ_DELTA_MAPPING = {
    "YS": (1, "Y" ), "Y": (1, "Y" ), "AS": (1, "Y" ), "A": (1, "Y" ),
    "QS": (3, "M" ), "Q": (3, "M" ),
    "MS": (1, "M" ), "M": (1, "M" ),
    "W": (7, "D" ),
    "D": (1, "D" ),
    "H": (1, "H" ),
}

# Special CF unit
PART_OF_A_WHOLE_UNIT = "1"

# companion parameter to slice_mode
GROUP_BY_METHOD = "groupby"
RESAMPLE_METHOD = "resample"
GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD = "groupby_ref_and_resample_study"

DEFAULT_DOY_WINDOW = 5



from __future__ import annotations

import os
from datetime import datetime
from unittest.mock import MagicMock, patch

import cftime
import numpy as np
import pandas as pd
import pint
import pytest
import xarray as xr

import icclim
from icclim.ecad.ecad_indices import EcadIndexRegistry
from icclim.generic_indices.threshold import build_threshold
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import (
    ICCLIM_VERSION,
    PART_OF_A_WHOLE_UNIT,
    REFERENCE_PERIOD_ID,
    UNITS_KEY,
)
from icclim.models.frequency import FrequencyRegistry
from icclim.models.index_group import IndexGroupRegistry
from icclim.tests.testing_utils import K2C, stub_pr, stub_tas


@patch("icclim.main.index")
@patch("icclim.icclim_logger.IcclimLogger")
def test_deprecated_indice(log_mock: MagicMock, index_mock: MagicMock):
    icclim.main.log = log_mock
    icclim.indice()
    log_mock.deprecation_warning.assert_called_once_with(
        old="icclim.indice", new="icclim.index"
    )
    index_mock.assert_called_once()


HEAT_INDICES = ["SU", "TR", "WSDI", "TG90p", "TN90p", "TX90p", "TXx", "TNx", "CSU"]


@pytest.mark.slow
class Test_Integration:
    """
    Integration tests.
    We are not testing here the actual indices results, they are already tested in
    `test_ecad_indices.py` as well as in xclim directly.
    The goal it to make sure every the whole app can run smoothly

    These tests have side effect:
    - writing and removing of "out.nc" file

    """

    OUTPUT_FILE = "out.nc"
    TIME_RANGE = pd.date_range(start="2042-01-01", end="2045-12-31", freq="D")
    CF_TIME_RANGE = xr.cftime_range("2042-01-01", end="2045-12-31", freq="D")
    data = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(lat=[42], lon=[42], time=TIME_RANGE),
        attrs={UNITS_KEY: "degC"},
    )
    full_data = data.to_dataset(name="tas")
    full_data["tasmax"] = data
    full_data["tasmin"] = data
    full_data["pr"] = data.copy(deep=True)
    full_data["pr"].attrs[UNITS_KEY] = "kg m-2 d-1"
    full_data["snd"] = data.copy(deep=True)
    full_data["snd"].attrs[UNITS_KEY] = "cm"
    full_data["DD"] = data.copy(deep=True)
    full_data["DD"].attrs[UNITS_KEY] = "degree"
    full_data["wsgs_max"] = data.copy(deep=True)
    full_data["wsgs_max"].attrs[UNITS_KEY] = "m/s"
    full_data["sfcWind"] = data.copy(deep=True)
    full_data["sfcWind"].attrs[UNITS_KEY] = "m/s"

    data_cf_time = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(lat=[42], lon=[42], time=CF_TIME_RANGE),
        attrs={UNITS_KEY: "degC"},
    )

    # usually, time_bounds is not properly decoded an keep a object dtype
    time_bounds = xr.DataArray(
        data=[[t, t + np.timedelta64(1, "h")] for t in data.time.values],
        dims=["time", "bounds"],
        coords=dict(bounds=[0, 1], time=TIME_RANGE),
    ).astype("object")

    dataset_with_time_bounds = xr.Dataset(
        dict(data=data, time_bounds=time_bounds),
    )

    not_spi_indices = list(
        filter(lambda x: "spi" not in x.short_name.lower(), EcadIndexRegistry.values())
    )

    @pytest.fixture(autouse=True)
    def cleanup(self):
        # setup
        # ...
        yield
        # teardown
        try:
            os.remove(self.OUTPUT_FILE)
        except FileNotFoundError:
            pass

    def test_index_SU(self):
        tas = stub_tas(tas_value=26 + K2C)
        tas[:5] = 0
        res = icclim.index(
            index_name="SU", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        assert f"icclim version: {ICCLIM_VERSION}" in res.attrs["history"]
        assert res.SU.isel(time=0) == 26  # January

    def test_index_SU__on_dataset(self):
        res = icclim.index(
            index_name="SU",
            var_name="data",
            in_files=self.dataset_with_time_bounds,
            out_file=self.OUTPUT_FILE,
        )
        assert f"icclim version: {ICCLIM_VERSION}" in res.attrs["history"]
        np.testing.assert_array_equal(0, res.SU)

    def test_index_DTR(self):
        ds = self.data.to_dataset(name="toto")
        ds["tutu"] = self.data
        res = icclim.index(
            index_name="DTR",
            in_files=ds,
            out_file=self.OUTPUT_FILE,
            var_name=["toto", "tutu"],
        )
        assert f"icclim version: {ICCLIM_VERSION}" in res.attrs["history"]
        np.testing.assert_array_equal(0, res.DTR)

    def test_index_CD(self):
        ds = self.data.to_dataset(name="tas")
        ds["pr"] = self.data.copy(deep=True)
        ds["pr"].attrs[UNITS_KEY] = "kg m-2 d-1"
        res = icclim.index(
            index_name="CD",
            in_files=ds,
            out_file=self.OUTPUT_FILE,
        )
        assert f"icclim version: {ICCLIM_VERSION}" in res.attrs["history"]
        np.testing.assert_array_equal(0, res.CD)

    def test__preserve_initial_history(self):
        self.data.attrs["history"] = "pouet pouet cacahuête"
        res = icclim.su(in_files=self.data)
        assert "pouet pouet cacahuête" in res.attrs["history"]

    def test_index_SU__time_selection(self):
        # WHEN
        res_string_dates = icclim.index(
            index_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            time_range=("19 july 2042", "14 august 2044"),
        )
        res_datetime_dates = icclim.index(
            index_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            time_range=[datetime(2042, 7, 19), datetime(2044, 8, 14)],
        )
        # THEN
        assert res_string_dates.time_bounds[0, 0] == np.datetime64(datetime(2042, 1, 1))
        assert res_string_dates.time_bounds[0, 1] == np.datetime64(
            datetime(2042, 12, 31)
        )
        np.testing.assert_array_equal(res_string_dates.SU, res_datetime_dates.SU)
        np.testing.assert_array_equal(
            res_string_dates.time_bounds, res_datetime_dates.time_bounds
        )

    def test_index_SU__pandas_time_slice_mode(self):
        # WHEN
        res = icclim.index(
            index_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            slice_mode="2W-WED",
        )
        # THEN
        assert res.time_bounds[0, 0] == np.datetime64(datetime(2042, 1, 1))
        assert res.time_bounds[0, 1] == np.datetime64(datetime(2042, 1, 14))
        assert (
            res.SU.attrs["standard_name"]
            == "number_of_days_when_maximum_air_temperature_is_greater_than_threshold"
        )
        assert (
            res.SU.attrs["long_name"]
            == "Number of days when maximum air temperature is greater than 25.0"
            " degC for each 2 wednesday starting week(s)."
        )

    def test_index_SU__monthy_sampled(self):
        res = icclim.index(
            index_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            slice_mode=FrequencyRegistry.MONTH,
        )
        np.testing.assert_array_equal(0, res.SU)
        np.testing.assert_array_equal(
            len(np.unique(self.TIME_RANGE.year)) * 12, len(res.time)
        )

    def test_index_SU__monthy_sampled_cf_time(self):
        res = icclim.index(
            index_name="SU",
            in_files=self.data_cf_time,
            out_file=self.OUTPUT_FILE,
            slice_mode=FrequencyRegistry.MONTH,
        )
        np.testing.assert_array_equal(0, res.SU)
        np.testing.assert_array_equal(
            len(np.unique(self.TIME_RANGE.year)) * 12, len(res.time)
        )
        assert res.time_bounds.sel(time=res.time[0])[0] == cftime.DatetimeGregorian(
            2042, 1, 1, 0, 0, 0, 0
        )
        assert res.time_bounds.sel(time=res.time[0])[1] == cftime.DatetimeGregorian(
            2042, 1, 31, 0, 0, 0, 0
        )

    def test_index_SU__DJF_cf_time(self):
        res = icclim.index(
            index_name="SU",
            in_files=self.data_cf_time,
            out_file=self.OUTPUT_FILE,
            slice_mode=FrequencyRegistry.DJF,
        )
        np.testing.assert_array_equal(res.SU.isel(time=0), np.NAN)
        np.testing.assert_array_equal(res.SU.isel(time=1), 0)
        # "+ 1" because DJF sampling create a december month with nans before first year
        np.testing.assert_array_equal(
            len(np.unique(self.TIME_RANGE.year)) + 1, len(res.time)
        )
        assert res.time_bounds.sel(time=res.time[0])[0] == cftime.DatetimeGregorian(
            2041, 12, 1, 0, 0, 0, 0
        )
        assert res.time_bounds.sel(time=res.time[0])[1] == cftime.DatetimeGregorian(
            2042, 2, 28, 0, 0, 0, 0
        )

    def test_indices__from_DataArray(self):
        res = icclim.indices(
            index_group=IndexGroupRegistry.HEAT,
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
        )
        for i in HEAT_INDICES:
            assert res[i] is not None

    def test_indices__on_var_name(self):
        res = icclim.indices(
            index_group="tasmax",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
        )
        for i in ["SU", "WSDI", "TX90p", "TXx", "CSU", "ID", "TX10p", "TXn"]:
            assert res[i] is not None

    def test_indices__on_index_name(self):
        res = icclim.indices(
            index_group="tx90p",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
        )
        for i in ["TX90p"]:
            assert res[i] is not None

    def test_indices__on_var_names(self):
        ds = self.data.to_dataset(name="tas")
        ds["pr"] = self.data.copy(deep=True)
        ds["pr"].attrs[UNITS_KEY] = "kg m-2 d-1"
        res = icclim.indices(
            index_group=["tas", "pr"],
            in_files=ds,
            out_file=self.OUTPUT_FILE,
            base_period_time_range=("2042-01-01", "2042-12-31"),
        )
        for i in [
            "TG90p",
            "GD4",
            "HD17",
            "TG10p",
            "CDD",
            "PRCPTOT",
            "RR1",
            "SDII",
            "CWD",
            "RR",
            "R10mm",
            "R20mm",
            "RX1day",
            "RX5day",
            "R75p",
            "R75pTOT",
            "R95p",
            "R95pTOT",
            "R99p",
            "R99pTOT",
            "CD",
            "CW",
            "WD",
            "WW",
        ]:
            assert res[i] is not None

    def test_indices__on_group_union(self):
        ds = self.data.to_dataset(name="tx")
        ds["tn"] = self.data.copy(deep=True)
        ds["tg"] = self.data.copy(deep=True)
        ds["snd"] = self.data.copy(deep=True)
        ds["snd"].attrs[UNITS_KEY] = "mm"
        res = icclim.indices(
            index_group=IndexGroupRegistry.HEAT | IndexGroupRegistry.SNOW,
            in_files=ds,
            out_file=self.OUTPUT_FILE,
        )
        for i in [
            "SU",
            "TR",
            "WSDI",
            "TG90p",
            "TN90p",
            "TX90p",
            "TXx",
            "TNx",
            "CSU",
            "SD",
            "SD1",
            "SD5cm",
            "SD50cm",
        ]:
            assert res[i] is not None

    def test_indices__error(self):
        ds = self.data.to_dataset(name="tx")
        with pytest.raises(InvalidIcclimArgumentError):
            icclim.indices(
                index_group="wubaluba dub dub",
                in_files=ds,
                out_file=self.OUTPUT_FILE,
            )

    def test_indices__snow_indices(self):
        ds = self.data.to_dataset(name="tas")
        ds["snd"] = self.data.copy(deep=True)
        ds["snd"].attrs[UNITS_KEY] = "cm"
        res = icclim.indices(
            index_group=IndexGroupRegistry.SNOW, in_files=ds, out_file=self.OUTPUT_FILE
        )
        for i in filter(
            lambda i: i.group == IndexGroupRegistry.SNOW, EcadIndexRegistry.values()
        ):
            assert res[i.short_name] is not None

    def test_indices_all_from_Dataset(self):
        res = icclim.indices(
            index_group="all",
            in_files=self.full_data,
            out_file=self.OUTPUT_FILE,
            base_period_time_range=("2042-01-01", "2042-12-31"),
        )
        for i in EcadIndexRegistry.values():
            assert res[i.short_name] is not None

    def test_indices_all_from_Dataset__seasonal_SPI_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            icclim.indices(
                index_group="SPI3",
                in_files=self.full_data,
                out_file=self.OUTPUT_FILE,
                slice_mode=["season", [1, 2, 3]],
                base_period_time_range=("2042-01-01", "2042-12-31"),
            )

    def test_indices_all_from_Dataset__seasonal(self):
        res = icclim.indices(
            index_group="all",
            in_files=self.full_data,
            out_file=self.OUTPUT_FILE,
            slice_mode=["season", [1, 2, 3]],
            ignore_error=True,
        )
        for i in self.not_spi_indices:
            assert res[i.short_name] is not None

    def test_indices_all_from_Dataset__between_dates_seasonal(self):
        res = icclim.indices(
            index_group="all",
            in_files=self.full_data,
            out_file=self.OUTPUT_FILE,
            slice_mode=["season", ["07-19", "08-14"]],
            ignore_error=True,
        )
        for i in self.not_spi_indices:
            assert res[i.short_name] is not None

    def test_indices_all_from_Dataset__JFM_seasonal(self):
        res = icclim.indices(
            index_group="all",
            in_files=self.full_data,
            out_file=self.OUTPUT_FILE,
            slice_mode=["season", [1, 2, 3]],
            ignore_error=True,
        )
        for i in self.not_spi_indices:
            assert res[i.short_name] is not None

    def test_indices_all_from_Dataset__between_year_season(self):
        res = icclim.indices(
            index_group="all",
            in_files=self.full_data,
            out_file=self.OUTPUT_FILE,
            slice_mode=["season", [12, 1, 2, 3]],
            ignore_error=True,
        )
        for i in self.not_spi_indices:
            assert res[i.short_name] is not None

    def test_indices_all_ignore_error(self):
        no_snow = self.full_data.copy()
        del no_snow["snd"]
        res: xr.Dataset = icclim.indices(
            index_group="all",
            in_files=no_snow,
            out_file=self.OUTPUT_FILE,
            ignore_error=True,
            slice_mode="DJF",
            base_period_time_range=("2042-01-01", "2042-12-31"),
        ).compute()
        for i in EcadIndexRegistry.values():
            # No variable in input to compute snow indices
            if i.group == IndexGroupRegistry.SNOW:
                assert res.data_vars.get(i.short_name, None) is None
            elif "spi" in i.short_name.lower():
                assert res.data_vars.get(i.short_name, None) is None
            else:
                assert res[i.short_name] is not None

    def test_indices_all__error(self):
        ds = self.data.to_dataset(name="tas")
        ds["tasmax"] = self.data
        ds["tasmin"] = self.data
        ds["pr"] = self.data.copy(deep=True)
        ds["pr"].attrs[UNITS_KEY] = "kg m-2 d-1"
        with pytest.raises(Exception):
            icclim.indices(
                index_group="all",
                in_files=ds,
                out_file=self.OUTPUT_FILE,
                ignore_error=False,
            )

    def test_index_TR(self):
        tas = stub_tas(tas_value=26 + K2C)
        tas[:5] = 0
        res = icclim.index(
            index_name="TR", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        assert f"icclim version: {ICCLIM_VERSION}" in res.attrs["history"]
        assert res.TR.isel(time=0) == 26  # January

    def test_index_prcptot(self):
        pr = stub_pr(value=2)
        pr[:10] = 0
        res = icclim.index(
            index_name="prcptot",
            in_files=pr,
            out_file=self.OUTPUT_FILE,
            slice_mode="ms",
        )
        assert res.isel(time=0) == 42.0

    def test_index_csu(self):
        tas = stub_tas(tas_value=26 + K2C)
        tas[10:15] = 0
        res = icclim.index(
            index_name="csu", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        assert res.isel(time=0) == 16

    def test_index_gd4(self):
        tas = stub_tas(tas_value=26 + K2C)
        tas[5:15] = 0
        res = icclim.index(
            index_name="gd4", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        expected = (26 - 4) * 21
        assert (
            res.isel(time=0) == expected
        )  # 21 days in January above 4 degC (at 26degC)

    def test_index_cfd(self):
        tas = stub_tas(tas_value=26 + K2C)
        tas[5:15] = 0
        res = icclim.index(
            index_name="cfd", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        assert res.isel(time=0) == 1

    def test_index_fd(self):
        tas = stub_tas(tas_value=26 + K2C)
        tas[5:15] = 0
        tas[20:25] = 0
        res = icclim.index(
            index_name="fd", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        assert res.isel(time=0) == 15

    def test_index_hd17(self):
        tas = stub_tas(tas_value=27 + K2C)
        tas[5:10] = 0
        res = icclim.index(
            index_name="hd17", in_files=tas, out_file=self.OUTPUT_FILE, slice_mode="ms"
        )
        assert res.isel(time=0) == 5 * (17 + K2C)

    def test_index_tx90p__no_bootstrap_because_one_single_year_of_ref(self):
        tas = stub_tas(tas_value=27 + K2C)
        tas[5:15] = 0
        res = icclim.index(
            index_name="tx90p",
            in_files=tas,
            doy_window_width=5,
            base_period_time_range=("2042-01-01", "2042-12-31"),
            time_range=("2042-01-01", "2045-12-31"),
            out_file=self.OUTPUT_FILE,
            slice_mode="ms",
        )
        assert REFERENCE_PERIOD_ID not in res.TX90p.attrs
        # The 90th percentile here is clipped to the maximum of tas window (27 degC)
        # due to the "median_unbiased" interpolation.
        # Thus no value are strictly above it.
        assert res.TX90p.isel(time=0) == 0

    def test_index_tx90p__no_bootstrap_because_no_overlap(self):
        tas = stub_tas(tas_value=27 + K2C)
        tas[5:10] = 0
        res = icclim.index(
            index_name="tx90p",
            in_files=tas,
            doy_window_width=1,
            time_range=("2043-01-01", "2045-12-31"),
            base_period_time_range=("2042-01-01", "2042-12-31"),
            out_file=self.OUTPUT_FILE,
            slice_mode="ms",
        )
        assert REFERENCE_PERIOD_ID not in res.TX90p.attrs
        # resample_doy add a day where 90th per is below tas
        assert res.TX90p.isel(time=0) == 6

    def test_index_tx90p__bootstrap_2_years(self):
        tas = stub_tas(tas_value=27 + K2C)
        tas[5:10] = 0
        res = icclim.index(
            index_name="tx90p",
            in_files=tas,
            doy_window_width=1,
            time_range=("2042-01-01", "2045-12-31"),
            base_period_time_range=("2042-01-01", "2043-12-31"),
            out_file=self.OUTPUT_FILE,
            slice_mode="ms",
        )
        assert REFERENCE_PERIOD_ID in res.TX90p.attrs
        # 2042 values are compared to 2043's 90th percentile due to bootstrap
        assert res.TX90p.sel(time="2042-01") == 0
        # 2043 values are compared to 2042's 90th percentile due to bootstrap
        assert res.TX90p.sel(time="2043-01") == 5

    def test_index_wsdi__no_bootstrap_because_no_overlap(self):
        tas = stub_tas(tas_value=27 + K2C)
        tas[0:10] = 0
        res = icclim.index(
            index_name="wsdi",
            in_files=tas,
            doy_window_width=1,
            time_range=("2043-01-01", "2045-12-31"),
            base_period_time_range=("2042-01-01", "2042-12-31"),
            out_file=self.OUTPUT_FILE,
            slice_mode="ms",
        )
        assert REFERENCE_PERIOD_ID not in res.WSDI.attrs
        # 1 more day than in tas because of resample_doy that interpolate values
        assert res.WSDI.isel(time=0) == 11

    def test_index_csdi__no_bootstrap_because_no_overlap(self):
        tas = stub_tas(tas_value=2 + K2C)
        tas[0:10] = 35 + K2C
        res = icclim.index(
            index_name="csdi",
            in_files=tas,
            doy_window_width=1,
            time_range=("2043-01-01", "2045-12-31"),
            base_period_time_range=("2042-01-01", "2042-12-31"),
            out_file=self.OUTPUT_FILE,
            slice_mode="ms",
        )
        assert REFERENCE_PERIOD_ID not in res.CSDI.attrs
        # 1 more day than in tas because of resample_doy that interpolate values
        assert res.CSDI.isel(time=0) == 11

    def test_count_occurrences__date_event(self):
        tas = stub_tas(tas_value=2 + K2C)
        tas[10] = 35 + K2C
        res = icclim.index(
            tas,
            var_name=["tmin"],
            index_name="count_occurrences",
            threshold=">= 22 degree_Celsius",
            slice_mode="month",
            date_event=True,
        ).compute()
        assert "event_date_start" in res.coords
        assert "event_date_end" in res.coords
        assert res.count_occurrences.isel(time=0).event_date_end == np.datetime64(
            "2042-01-11"
        )
        assert res.count_occurrences.isel(time=0).event_date_start == np.datetime64(
            "2042-01-11"
        )

    def test_count_occurrences__to_percent(self):
        tas = stub_tas(tas_value=2 + K2C)
        tas[10] = 35 + K2C
        res = icclim.index(
            tas,
            var_name=["tmin"],
            index_name="count_occurrences",
            threshold=">= 22 degree_Celsius",
            slice_mode="month",
            out_unit="%",
        ).compute()
        assert res.count_occurrences.attrs[UNITS_KEY] == "%"
        assert res.count_occurrences.isel(time=0) == 1 / 31 * 100

    def test_count_occurrences__multiple_simple_thresholds(self):
        tas = stub_tas(tas_value=2 + K2C)
        tas[10] = 35 + K2C
        res = icclim.index(
            tas,
            var_name=["tmin"],
            index_name="count_occurrences",
            threshold=build_threshold(value=[1, 30], operator=">=", unit="deg_C"),
            slice_mode="month",
            save_thresholds=True,
        ).compute()
        assert res.count_occurrences.attrs[UNITS_KEY] == "d"
        assert res.count_occurrences.isel(time=0).sel(threshold=1) == 31
        # The 5 days rolling turn the 1 day unusual value into a 5 day time lapse
        assert res.count_occurrences.isel(time=0).sel(threshold=30) == 1

    def test_count_occurrences__multiple_doy_per_thresholds(self):
        tas = stub_tas(tas_value=2 + K2C)
        tas[10] = 35 + K2C
        res = icclim.index(
            tas,
            var_name=["tmin"],
            index_name="count_occurrences",
            threshold=build_threshold(value=[10, 99], operator=">=", unit="doy_per"),
            slice_mode="month",
            save_thresholds=True,
        ).compute()
        assert res.count_occurrences.attrs[UNITS_KEY] == "d"
        assert res.count_occurrences.isel(time=0).sel(percentiles=10) == 31
        # The 5 days rolling turn the 1 day unusual value into a 5 day time lapse
        assert res.count_occurrences.isel(time=0).sel(percentiles=99) == 26

    def test_count_occurrences__multiple_period_per_thresholds(self):
        tas = stub_tas(tas_value=-20 + K2C)
        tas[10] = 35 + K2C
        res = icclim.index(
            tas,
            var_name=["tmin"],
            index_name="count_occurrences",
            threshold=build_threshold(
                value=[10, 99.95], operator=">=", unit="period_per"
            ),
            slice_mode="month",
            save_thresholds=True,
        ).compute()
        assert res.count_occurrences.attrs[UNITS_KEY] == "d"
        assert res.count_occurrences.isel(time=0).sel(percentiles=10) == 31
        assert res.count_occurrences.isel(time=0).sel(percentiles=99.95) == 1

    def test_excess__on_doy_percentile(self):
        tas = stub_tas(tas_value=10 + K2C).rename("tas")
        tas[10] = 5 + K2C
        res = icclim.index(
            tas,
            index_name="excess",
            time_range=["2044-01-01", "2045-12-31"],
            threshold=build_threshold(
                "10 doy_per",
                doy_window_width=1,
                reference_period=["2042-01-01", "2042-12-31"],
            ),
            slice_mode="month",
            save_thresholds=True,
        ).compute()
        # not exactly 5 because of resample_doy interpolation
        np.testing.assert_almost_equal(res.excess.isel(time=0), 5.01369863)
        assert "tas_thresholds" in res.data_vars

    def test_deficit__on_doy_percentile(self):
        tas = stub_tas(tas_value=5 + K2C).rename("tas")
        tas[10] = 10 + K2C
        res = icclim.index(
            tas,
            index_name="deficit",
            time_range=["2044-01-01", "2045-12-31"],
            threshold=build_threshold(
                "10 doy_per",
                doy_window_width=1,
                reference_period=["2042-01-01", "2042-12-31"],
            ),
            slice_mode="month",
            save_thresholds=True,
        ).compute()
        # not exactly 5 because of resample_doy interpolation
        np.testing.assert_almost_equal(res.deficit.isel(time=0), 5.01369863)
        assert "tas_thresholds" in res.data_vars

    def test_fraction_of_total(self):
        tas = stub_tas(tas_value=25 + K2C).rename("tas")
        tas[tas.time.dt.date == np.datetime64("2042-06-10")] = 10 + K2C
        res = icclim.index(
            tas,
            index_name="fraction_of_total",
            threshold="> 20 degree_Celsius",
            slice_mode="jja",
        ).compute()
        np.testing.assert_almost_equal(res.fraction_of_total.isel(time=0), 0.98967164)
        assert res.fraction_of_total.isel(time=1) == 1
        assert res.fraction_of_total.attrs[UNITS_KEY] == PART_OF_A_WHOLE_UNIT

    def test_fraction_of_total_percent(self):
        tas = stub_tas(tas_value=25 + K2C).rename("tas")
        tas[tas.time.dt.date == np.datetime64("2042-06-10")] = 10 + K2C
        res = icclim.index(
            tas,
            index_name="fraction_of_total",
            threshold="> 20 degree_Celsius",
            out_unit="%",
            slice_mode="jja",
        ).compute()
        np.testing.assert_almost_equal(res.fraction_of_total.isel(time=0), 98.96716372)
        assert res.fraction_of_total.isel(time=1) == 100
        assert res.fraction_of_total.attrs[UNITS_KEY] == "%"

    def test_std(self):
        tas = stub_tas(tas_value=25 + K2C).rename("tas")
        res = icclim.index(
            tas,
            index_name="standard_deviation",
        ).compute()
        np.testing.assert_almost_equal(res.standard_deviation.isel(time=0), 0)

    def test_slice_mode__between_date(self):
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        precipitation = xr.DataArray(
            np.ones(365),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "mm/day"},
        )
        precipitation[0:5] = [0.1, 0.1, 0.1, 2, 3]
        cdd = icclim.cdd(
            in_files=precipitation, slice_mode=["season", ["01-02", "01-05"]]
        ).CDD
        # The 01-01 value is ignored because we clip the wanted season before computing
        # the index
        np.testing.assert_almost_equal(cdd.isel(time=0), 2)

    def test_rr_with_slice_mode__week(self):
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        precipitation = xr.DataArray(
            np.zeros(365),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "mm/day"},
        )
        precipitation[0:5] = [0.1, 0.1, 0.1, 2, 3]
        rr = icclim.rr(in_files=precipitation, slice_mode="W").RR
        # The 01-01 value is ignored because we clip the wanted season before computing
        # the index
        np.testing.assert_almost_equal(rr.isel(time=0), 0.2)
        np.testing.assert_almost_equal(rr.isel(time=1), 5.1)

    def test_rr_with_slice_mode__4_weeks(self):
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        precipitation = xr.DataArray(
            np.zeros(365),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "mm/day"},
        )
        precipitation[0:5] = [0.1, 0.1, 0.1, 2, 3]
        rr = icclim.rr(in_files=precipitation, slice_mode="2W-FRI")
        # The 01-01 value is ignored because we clip the wanted season before computing
        # the index
        np.testing.assert_almost_equal(rr.RR.isel(time=0), 5.3)
        np.testing.assert_almost_equal(rr.RR.isel(time=1), 0)

    def test_mm_to_mmday(self):
        # GIVEN
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        precip = xr.DataArray(
            np.ones(365),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "mm", "standard_name": "thickness_of_rainfall_amount"},
        )
        precip.loc[{"time": slice("2000-01-01", "2000-01-05")}] = 50
        # WHEN
        r10mm = icclim.r10mm(in_files=precip, slice_mode="month")
        # THEN
        assert r10mm.isel(time=0) == 5

    def test_mm_to_mmday__error_bas_standard_name(self):
        # GIVEN
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        precip = xr.DataArray(
            np.ones(365),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "mm", "standard_name": "HeHoCacao"},
        )
        precip.loc[{"time": slice("2000-01-01", "2000-01-05")}] = 50
        # THEN
        with pytest.raises(pint.DimensionalityError):
            # WHEN
            icclim.r10mm(in_files=precip)

    def test_ddnorth(self):
        # GIVEN
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        dd = xr.DataArray(
            np.full(365, 300),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "degree"},
        )
        dd.loc[{"time": slice("2000-01-01", "2000-01-05")}] = 20  # north
        dd.loc[{"time": slice("2000-01-06", "2000-01-10")}] = 320  # north
        dd.loc[{"time": slice("2000-01-11", "2000-01-31")}] = 60  # east
        # WHEN
        ddnorth = icclim.ddnorth(in_files=dd, slice_mode="month").DDnorth.compute()
        # THEN
        np.testing.assert_almost_equal(ddnorth.isel(time=0), 10)

    def test_ddeast(self):
        # GIVEN
        time_range = xr.DataArray(
            pd.date_range("2000", periods=365, freq="D"), dims=["time"]
        )
        dd = xr.DataArray(
            np.full(365, 300),
            coords={"time": time_range, "lat": 1, "lon": 1},
            dims="time",
            attrs={"units": "degree"},
        )
        dd.loc[{"time": slice("2000-01-01", "2000-01-05")}] = 20  # north
        dd.loc[{"time": slice("2000-01-06", "2000-01-10")}] = 320  # north
        dd.loc[{"time": slice("2000-01-11", "2000-01-31")}] = 60  # east
        # WHEN
        ddeast = icclim.ddeast(in_files=dd, slice_mode="month").DDeast.compute()
        # THEN
        np.testing.assert_almost_equal(ddeast.isel(time=0), 21)
